import logging
from uuid import UUID

import orjson
from beartype import beartype
from deprecation import deprecated

from picsellia.colors import Colors
from picsellia.decorators import exception_handler
from picsellia.exceptions import ResourceNotFoundError
from picsellia.sdk.connexion import Connection
from picsellia.sdk.dao import Dao
from picsellia.sdk.dataset_version import DatasetVersion
from picsellia.sdk.resource import Resource
from picsellia.sdk.tag import Tag, TagTarget
from picsellia.sdk.taggable import Taggable
from picsellia.sdk.worker import Worker
from picsellia.types.enums import InferenceType, WorkerType
from picsellia.types.schemas import DatasetSchema
from picsellia.utils import filter_payload

logger = logging.getLogger("picsellia")


class Dataset(Dao, Taggable, Resource):
    def __init__(self, connexion: Connection, data: dict):
        Dao.__init__(self, connexion, data)
        Taggable.__init__(self, TagTarget.DATASET)
        Resource.__init__(self, "dataset")

    @property
    def name(self) -> str:
        """Name of this (Dataset)"""
        return self._name

    def __str__(self):
        return f"{Colors.YELLOW}Dataset {self.name} {Colors.ENDC} (id: {self.id})"

    @exception_handler
    @beartype
    def refresh(self, data: dict):
        schema = DatasetSchema(**data)
        self._name = schema.name
        return schema

    @exception_handler
    @beartype
    def sync(self) -> dict:
        r = self.connexion.get(f"/api/dataset/{self.id}").json()
        self.refresh(r)
        return r

    @exception_handler
    @beartype
    def get_tags(self) -> list[Tag]:
        """Retrieve the tags of your dataset.

        Examples:
            ```python
            tags = foo_dataset.get_tags()
            assert tags[0].name == "my-dataset-1"
            ```

        Returns:
            List of (Tag) objects
        """
        r = self.sync()
        return [Tag(self.connexion, item) for item in r["tags"]]

    @exception_handler
    @beartype
    def get_resource_url_on_platform(self) -> str:
        """Get platform url of this resource.

        Examples:
            ```python
            print(foo_dataset.get_resource_url_on_platform())
            >>> "https://app.picsellia.com/dataset/62cffb84-b92c-450c-bc37-8c4dd4d0f590"
            ```

        Returns:
            Url on Platform for this resource
        """

        return f"{self.connexion.host}/dataset/{self.id}"

    @exception_handler
    @beartype
    def delete(self) -> None:
        """Delete a dataset.

        :warning: **DANGER ZONE**: Be very careful here!

        It will remove this dataset from our database, its versions with their assets and annotations will be removed.
        It will also remove potential annotation campaigns of this dataset versions.

        Examples:
            ```python
            foo_dataset.delete()
            ```
        """
        self.connexion.delete(f"/api/dataset/{self.id}")
        logger.info(f"{self} deleted")

    @exception_handler
    @beartype
    def update(
        self,
        name: str | None = None,
        private: bool | None = None,
        description: str | None = None,
    ) -> None:
        """Update name, private or description of this Dataset.

        Examples:
            ```python
            dataset.update(description='My favourite dataset')
            ```

        Arguments:
            name (str, optional): New name of the dataset. Defaults to None.
            private (bool, optional): New private of the dataset. Defaults to None.
            description (str, optional): New description of the dataset. Defaults to None.
        """
        if private is not None:
            logging.warning(
                "'private' parameter is deprecated and will be removed in future versions. "
                "You cannot update to a public dataset from the SDK anymore."
            )

        payload = {"name": name, "description": description}
        filtered_payload = filter_payload(payload)
        r = self.connexion.patch(
            f"/api/dataset/{self.id}", data=orjson.dumps(filtered_payload)
        ).json()
        self.refresh(r)
        logger.info(f"{self} updated")

    @exception_handler
    @beartype
    def list_versions(
        self,
        limit: int | None = None,
        offset: int | None = None,
        order_by: list[str] | None = None,
    ) -> list[DatasetVersion]:
        """List all versions of this dataset

        Examples:
            ```python
            dataset.list_versions()
            ```

        Arguments:
            limit (int, optional): limit of versions to retrieve. Defaults to None.
            offset (int, optional): offset to start retrieving versions. Defaults to None.
            order_by (list[str], optional): fields to order by. Defaults to None.

        Returns:
            List of (DatasetVersion) objects
        """
        params = {"limit": limit, "offset": offset, "order_by": order_by}
        params = filter_payload(params)
        r = self.connexion.get(f"/api/dataset/{self.id}/versions", params=params).json()
        return [DatasetVersion(self.connexion, item) for item in r["items"]]

    @exception_handler
    @beartype
    def get_version(self, version: str) -> DatasetVersion:
        """Retrieve one version of a dataset

        Examples:
            ```python
            my_dataset_version = my_dataset.get_version("first")
            ```

        Arguments:
            version (str): version name to retrieve

        Returns:
            a (DatasetVersion) object
        """
        params = {"version": version}
        r = self.connexion.get(
            f"/api/dataset/{self.id}/versions/find", params=params
        ).json()
        return DatasetVersion(self.connexion, r)

    @exception_handler
    @beartype
    def get_version_by_id(self, id: UUID | str) -> DatasetVersion:
        """Retrieve one version of a dataset

        Examples:
            ```python
            my_dataset_version = my_dataset.get_version_by_id("918351d2-3e96-4970-bb3b-420f33ded895")
            ```

        Arguments:
            id (UUID): id of the version to retrieve

        Returns:
            a (DatasetVersion) object
        """
        if isinstance(id, str):
            id = UUID(id)
        params = {"id": id}
        r = self.connexion.get(
            f"/api/dataset/{self.id}/versions/find", params=params
        ).json()
        return DatasetVersion(self.connexion, r)

    @exception_handler
    @beartype
    def create_version(
        self,
        version: str,
        description: str = "",
        type: InferenceType | str = InferenceType.NOT_CONFIGURED,
    ) -> DatasetVersion:
        """Create a version of this dataset.

        A versioned dataset (DatasetVersion) takes (Data) from (Datalake) and transform it as annotable (Asset).

        Examples:
            ```python
            foo_dataset = client.create_dataset('foo_dataset')
            foo_dataset_version_1 = foo_dataset.create_version('first')
            some_data = client.get_datalake().list_data(limit=10)
            foo_dataset_version_1.add_data(some_data)
            ```

        Arguments:
            version (str): version name
            description (str): description of this version
            type (InferenceType): type of this version

        Returns:
            A (DatasetVersion) manipulable that can receive data
        """
        payload = {
            "version": version,
            "description": description,
            "type": InferenceType.validate(type),
        }

        r = self.connexion.post(
            f"/api/dataset/{self.id}/versions", data=orjson.dumps(payload)
        ).json()
        return DatasetVersion(self.connexion, r)

    @exception_handler
    @beartype
    @deprecated(
        deprecated_in="6.24.0",
        removed_in="6.27.0",
        details="This method should not be called anymore. Instead use list_users()",
    )
    def list_workers(self) -> list[Worker]:
        """List all workers of this dataset

        Examples:
            ```python
            dataset.list_workers()
            ```

        Returns:
            List of (Worker) objects
        """
        return [
            Worker(
                self.connexion,
                {"id": item.id, "username": item.username, "user_id": item.id},
                WorkerType.DATASET,
            )
            for item in self.list_users()
        ]

    @exception_handler
    @beartype
    @deprecated(
        deprecated_in="6.24.0",
        removed_in="6.27.0",
        details="This method should not be called anymore. Instead use list_users()",
    )
    def find_worker(self, username: str) -> Worker:
        """Find worker of this dataset from its username

        Examples:
            ```python
            dataset.find_worker("John")
            ```

        Arguments:
            username (str): username of the worker on the platform

        Returns:
            A (Worker) object
        """

        for user in self.list_users():
            if user.username == username:
                return Worker(
                    self.connexion,
                    {"id": user.id, "username": user.username, "user_id": user.id},
                    WorkerType.DATASET,
                )

        raise ResourceNotFoundError(
            f"There is no worker with username '{username}' in this dataset."
        )



import logging
import os
import shutil
import sys
import zipfile
from collections.abc import Callable
from pathlib import Path

from beartype import beartype

from picsellia import exceptions as exceptions
from picsellia.decorators import exception_handler

logger = logging.getLogger("picsellia")


print_section_handler = os.getenv("PICSELLIA_SDK_SECTION_HANDLER", "0")


def zip_file(path: str | Path):
    zip_filepath = os.path.splitext(path)[0] + ".zip"
    with zipfile.ZipFile(zip_filepath, "w", zipfile.ZIP_DEFLATED) as zipf:
        zipf.write(path, path)

    return zip_filepath


def zip_dir(path: str | Path):
    if not os.path.isdir(path):
        return zip_file(path)

    zip_filepath = os.path.join(os.path.dirname(path), os.path.basename(path) + ".zip")
    with zipfile.ZipFile(zip_filepath, "w", zipfile.ZIP_DEFLATED) as zipf:
        for filepath in os.listdir(path):
            zipf.write(os.path.join(path, filepath), filepath)

            if os.path.isdir(os.path.join(path, filepath)):
                for inside_file in os.listdir(os.path.join(path, filepath)):
                    zipf.write(
                        os.path.join(path, filepath, inside_file),
                        os.path.join(filepath, inside_file),
                    )

    return zip_filepath


@exception_handler
@beartype
def unzip(filename: str, target_dir: str | Path = "./"):
    shutil.unpack_archive(filename, target_dir)


@exception_handler
@beartype
def generate_requirements_json(requirements_path: str | Path) -> list[dict]:
    """Generate a json file with the requirements from the requirements.txt file

    Arguments:
        requirements_path ([str]): [absolute path to requirements.txt file]

    Raises:
        exceptions.ResourceNotFoundError: [Filepath does match]
        Exception: [Wrong requirements file]

    Returns:
        [dict]: {
            'requirements': []{
                'package': (str) package name,
                'version': (str) package version
            }
        }
    """
    requirements = []
    try:
        with open(requirements_path) as f:
            lines = f.readlines()
    except Exception:
        raise exceptions.ResourceNotFoundError(f"{requirements_path} does not exists")
    try:
        for line in lines:
            if line[0] != "#":
                try:
                    package, version = line.split("==")
                except Exception:
                    package, version = line, ""
                tmp = {"package": package.rstrip(), "version": version.rstrip()}
                requirements.append(tmp)
    except Exception:  # pragma: no cover
        raise Exception("Malformed requirements file")
    return requirements


def print_line_return():
    if print_section_handler == "1":
        sys.stdout.write("\n")


def print_start_section():
    if print_section_handler == "1":
        sys.stdout.write("-----")


def print_stop_section():
    if print_section_handler == "1":
        sys.stdout.write("--*--")


def print_start_chapter_name(name: str):
    if print_section_handler == "1":
        sys.stdout.write(f"--#--{name}")


def print_logging_buffer(length: int):
    if print_section_handler == "1":
        sys.stdout.write(f"--{length}--")


def filter_payload(payload: dict) -> dict:
    return {k: v for k, v in payload.items() if v is not None}


def str_to_payload(items):
    return ",".join([f'"{str(item)}"' for item in items])


def convert_tag_list_to_query_language(tags, intersect_tags: bool, prefix: str = ""):
    if tags is None:
        return None

    if not isinstance(tags, list):
        tags = [tags]
    computed_tags = []
    for tag in tags:
        if isinstance(tag, str):
            computed_tags.append(tag)
        else:
            # tag is a Tag, cannot be imported here
            computed_tags.append(tag.name)

    if not computed_tags:
        return None

    if intersect_tags:
        return " and ".join([f'{prefix}tags.name = "{tag}"' for tag in computed_tags])
    else:
        return prefix + "tags.name in (" + str_to_payload(computed_tags) + ")"


def combine_two_ql(q1, q2):
    if q1 is None:
        if q2 is None:
            return None
        else:
            return q2
    else:
        if q2 is None:
            return q1
        else:
            return q1 + " and " + q2


def chunk_list(items: list, chunk_size: int):
    return [(items[i : i + chunk_size], i) for i in range(0, len(items), chunk_size)]


def flatten(items: list[list]) -> list:
    return [item for sublist in items for item in sublist]


def flatten_dict(items: list[dict], filter_value: Callable = lambda _: True) -> dict:
    final_results = {}
    for res in items:
        for key, value in res.items():
            if key not in final_results:
                final_results[key] = value
            elif filter_value(value):
                final_results[key] += value
    return final_results



import math

from picsellia_annotations.coco import Annotation as COCOAnnotation
from picsellia_annotations.coco import Category as COCOCategory
from picsellia_annotations.coco import COCOFile
from picsellia_annotations.coco import Image as COCOImage

from picsellia.sdk.asset import MultiAsset


class COCOFileBuilder:
    def __init__(self, categories: dict[str, int]):
        self.categories = categories
        self.assets = {}
        self.coco_annotations = []
        self.coco_images = []
        self.coco_categories = []

    @staticmethod
    def prepare_categories(
        label_names: list[str],
        enforced_ordered_labels: list[str] | None = None,
    ) -> dict[str, int]:
        """This method returns a dict with name matching category id of labels.
        If enforced_ordered_labels is given, it will first put labels in given order, and then put remaining labels
        If not, it will be in the natural order (so created_at)
        """
        labels = {}
        if enforced_ordered_labels:
            for k in range(len(enforced_ordered_labels)):
                labels[enforced_ordered_labels[k]] = k

        for label in label_names:
            if not enforced_ordered_labels or label not in enforced_ordered_labels:
                labels[label] = len(labels)

        return labels

    def load_coco_annotations(self, loaded_annotations: dict[str, list]) -> list[str]:
        for asset_id, annotations in loaded_annotations.items():
            if len(annotations) == 0:
                continue

            # As load_annotations() return the annotations ordered by -created_at
            # we need to build shape for the first annotation of the list
            annotation = annotations[0]
            image_id = len(self.assets)
            self.assets[asset_id] = image_id
            self._build_coco_shapes(image_id, annotation)

        return list(self.assets.keys())

    def load_coco_images(self, loaded_assets: MultiAsset, use_id: bool = False):
        for asset in loaded_assets:
            asset_id = str(asset.id)
            if asset_id in self.assets:
                if use_id:
                    file_name = asset.id_with_extension
                else:
                    file_name = asset.filename
                coco_image = COCOImage(
                    id=self.assets[asset_id],
                    file_name=file_name,
                    width=asset.width,
                    height=asset.height,
                )
                self.coco_images.append(coco_image)

    def load_coco_categories(self, skeletons: dict[str, dict]):
        for label_name, k in self.categories.items():
            category = COCOCategory(id=k, name=label_name)
            if label_name in skeletons:
                category.skeleton = skeletons[label_name]["edges"]
                category.keypoints = skeletons[label_name]["vertices"]
            self.coco_categories.append(category)

        # Return categories by order of id
        self.coco_categories = sorted(self.coco_categories, key=lambda cat: cat.id)

    def build(self) -> COCOFile:
        return COCOFile(
            categories=self.coco_categories,
            images=self.coco_images,
            annotations=self.coco_annotations,
        )

    def _build_coco_shapes(self, image_id: int, annotation: dict):  # noqa: C901
        if "classifications" in annotation and annotation["classifications"]:
            for classification in annotation["classifications"]:
                category_id = self.categories[classification["label"]]
                coco_annotation = COCOAnnotation(
                    id=len(self.coco_annotations),
                    image_id=image_id,
                    category_id=category_id,
                    bbox=[],
                    segmentation=[],
                )
                self.coco_annotations.append(coco_annotation)

        if "rectangles" in annotation and annotation["rectangles"]:
            for rectangle in annotation["rectangles"]:
                category_id = self.categories[rectangle["label"]]
                if "area" in rectangle:
                    area = rectangle["area"]
                else:
                    area = float(rectangle["w"] * rectangle["h"])
                coco_annotation = COCOAnnotation(
                    id=len(self.coco_annotations),
                    image_id=image_id,
                    category_id=category_id,
                    bbox=[
                        rectangle["x"],
                        rectangle["y"],
                        rectangle["w"],
                        rectangle["h"],
                    ],
                    segmentation=[],
                    area=area,
                )
                self.coco_annotations.append(coco_annotation)

        if "polygons" in annotation and annotation["polygons"]:
            for polygon in annotation["polygons"]:
                if (
                    "polygon" not in polygon or not polygon["polygon"]
                ):  # pragma: no cover
                    continue
                category_id = self.categories[polygon["label"]]
                bbox, shape = COCOFileBuilder._compute_bbox_and_shape(
                    polygon["polygon"]
                )
                if "area" in polygon:
                    area = polygon["area"]
                else:
                    area = None
                coco_annotation = COCOAnnotation(
                    id=len(self.coco_annotations),
                    image_id=image_id,
                    category_id=category_id,
                    bbox=bbox,
                    segmentation=[shape],  # COCO needs an array around polygons
                    area=area,
                )
                self.coco_annotations.append(coco_annotation)

        if "keypoints" in annotation and annotation["keypoints"]:
            for keypoints in annotation["keypoints"]:
                if (
                    "keypoints" not in keypoints or not keypoints["keypoints"]
                ):  # pragma: no cover
                    continue
                category_id = self.categories[keypoints["label"]]
                bbox, num_keypoints, shape = (
                    COCOFileBuilder._compute_bbox_and_shape_keypoints(
                        keypoints["keypoints"]
                    )
                )
                coco_annotation = COCOAnnotation(
                    id=len(self.coco_annotations),
                    image_id=image_id,
                    category_id=category_id,
                    bbox=bbox,
                    keypoints=shape,
                    num_keypoints=num_keypoints,
                    area=1.0,
                )
                self.coco_annotations.append(coco_annotation)

    @staticmethod
    def _compute_bbox_and_shape(
        polygon: list[list[int | float]],
    ) -> tuple[list[int | float], list[int | float]]:
        x_min = math.inf
        x_max = -1
        y_min = math.inf
        y_max = -1

        segmentation = []
        for point in polygon:
            x = point[0]
            x_min = min(x, x_min)
            x_max = max(x, x_max)
            y = point[1]
            y_min = min(y, y_min)
            y_max = max(y, y_max)
            segmentation.append(x)
            segmentation.append(y)

        return [x_min, y_min, x_max - x_min, y_max - y_min], segmentation

    @staticmethod
    def _compute_bbox_and_shape_keypoints(
        keypoints: list[list[int | float]],
    ) -> tuple[list[int | float], int, list[int | float]]:
        x_min = math.inf
        x_max = -1
        y_min = math.inf
        y_max = -1

        num_keypoints = 0
        shape = []
        for point in keypoints:
            x = point[0]
            y = point[1]
            v = point[2]
            if v != 0:
                num_keypoints += 1
                x_min = min(x, x_min)
                x_max = max(x, x_max)
                y_min = min(y, y_min)
                y_max = max(y, y_max)

            shape.append(x)
            shape.append(y)
            shape.append(v)

        return [x_min, y_min, x_max - x_min, y_max - y_min], num_keypoints, shape



import abc
import logging
from typing import ClassVar
from uuid import UUID

from beartype import beartype
from orjson import orjson

from picsellia.decorators import exception_handler
from picsellia.sdk.dao import Dao
from picsellia.sdk.user import User
from picsellia.sdk.worker import Worker
from picsellia.services.campaign import parse_user_weights_to_assignees
from picsellia.types.enums import CampaignStepType
from picsellia.types.schemas import CampaignStepSchema
from picsellia.utils import filter_payload

logger = logging.getLogger("picsellia")


class AbstractCampaignStep(Dao):
    _base_path: ClassVar[str]

    @property
    @abc.abstractmethod
    def campaign_id(self) -> UUID:
        pass

    @property
    def name(self) -> str:
        """Name of the step"""
        return self._name

    @property
    def order(self) -> int:
        """Order of the step"""
        return self._order

    @property
    def type(self) -> CampaignStepType:
        """Type of the step"""
        return self._type

    @exception_handler
    @beartype
    def refresh(self, data: dict):
        schema = CampaignStepSchema(**data)
        self._name = schema.name
        self._order = schema.order
        self._type = schema.type
        return schema

    @exception_handler
    @beartype
    def sync(self) -> dict:
        r = self.connexion.get(f"{self._base_path}/{self.id}").json()
        self.refresh(r)
        return r

    @exception_handler
    @beartype
    def update(
        self,
        name: str | None = None,
        assignees: list[tuple[User | UUID | Worker, float | int]] | None = None,
        sample_rate: float | None = None,
    ) -> None:
        """Update this campaign step parameters.

        `assignees` changed signature in 6.24, we now except a User or a user UUID.
        Worker will still be accepted until 6.26

        Examples:
            ```python
            foo_step.update(name="another-name")
            ```

        Arguments:
            name (str, optional): name of the step. Defaults to None.
            assignees (list of tuple of (User or UUID), float, optional): Can be used to assign users to this step.
                Defaults to None.
            sample_rate (float, optional): sample rate of  this step. Defaults to None.
        """
        payload = {"name": name, "sample_rate": sample_rate}
        if assignees:
            payload["assignees"] = parse_user_weights_to_assignees(assignees)

        filtered_payload = filter_payload(payload)
        r = self.connexion.patch(
            f"{self._base_path}/{self.id}",
            data=orjson.dumps(filtered_payload),
        ).json()
        self.refresh(r)
        logger.info(f"{self} updated")

    @exception_handler
    @beartype
    def redistribute_assignments(self) -> None:
        """Distribute again all the assignments to the step assignees according to weights settings previously defined.
        Useful if you add workers to this step and want them to have as many tasks as the others.
        """
        self.connexion.post(f"{self._base_path}/{self.id}/assignments/redistribute")
        logger.info(f"{self} assignments have been redistributed")

    @exception_handler
    @beartype
    def delete(self, pending_assignments_destination_step_id: UUID | str) -> None:
        """Delete this step from the campaign.
        It will move all assignments from this step to given destination step id.

        :warning: **DANGER ZONE**: Be very careful here!

        Arguments:
            pending_assignments_destination_step_id (str or UUID): step on which deleted step assignments will be moved.
        """
        payload = {
            "destination_step_id_for_pending_assignments": pending_assignments_destination_step_id
        }
        self.connexion.delete(
            f"{self._base_path}/{self.id}", data=orjson.dumps(payload)
        )
        logger.info(
            f"Assignments of {self} have been moved to {pending_assignments_destination_step_id}."
        )
        logger.info(f"{self} have been deleted.")



from uuid import UUID

from beartype import beartype

from picsellia.colors import Colors
from picsellia.decorators import exception_handler
from picsellia.exceptions import ResourceNotFoundError
from picsellia.sdk.artifact import Artifact
from picsellia.sdk.connexion import Connection
from picsellia.sdk.dao import Dao
from picsellia.sdk.evaluation import MultiEvaluation
from picsellia.sdk.log import Log
from picsellia.services.training import TrainingService
from picsellia.types.schemas import FastTrainingSchema


class FastTraining(Dao):
    def __init__(
        self, connexion: Connection, dataset_version_id: UUID, data: dict
    ) -> None:
        data["id"] = data["experiment_id"]
        Dao.__init__(self, connexion, data)
        self.dataset_version_id = dataset_version_id

    @property
    def experiment_id(self):
        """Id of (Experiment) for this Fast Training"""
        return self._experiment_id

    @property
    def experiment_name(self) -> str:
        """Name of (Experiment) for this Fast Training"""
        return self._experiment_name

    def __str__(self):
        return f"{Colors.BLUE}Fast Training for Experiment '{self.experiment_name}' {Colors.ENDC} (id: {self.id})"

    @exception_handler
    @beartype
    def refresh(self, data: dict):
        schema = FastTrainingSchema(**data)
        self._status = schema.status
        self._experiment_id = schema.experiment_id
        self._experiment_name = schema.experiment_name
        self._job_id = schema.job_id
        self._job_run_id = schema.job_run_id
        return schema

    @exception_handler
    @beartype
    def sync(self) -> dict:
        items = self.connexion.get(
            f"/api/dataset/version/{self.dataset_version_id}/fast-training",
            params={"experiment_id": self.experiment_id},
        ).json()["trainings"]
        if len(items) != 1:
            raise ResourceNotFoundError("This training was not found")
        self.refresh(items[0])
        return items[0]

    @exception_handler
    @beartype
    def list_evaluations(
        self,
        limit: int | None = None,
        offset: int | None = None,
        q: str | None = None,
    ) -> MultiEvaluation:
        """List evaluations done in this (FastTraining).

         Arguments:
            limit (int, optional): if given, will limit the number of evaluations returned
            offset (int, optional): if given, will return evaluations that would have been returned
                                    after this offset in given order
            q (str, optional): if given, will try to apply query to filter evaluations

        Returns:
            A (MultiEvaluation)
        """
        return TrainingService(self.connexion, self.experiment_id).list_evaluations(
            limit, offset, q
        )

    @exception_handler
    @beartype
    def list_logs(self, limit: int | None = None, offset: int | None = None) -> list:
        """List everything that has been logged.

        List everything that has been logged to an experiment.

        Examples:
            ```python
            logs = experiment.list_logs()
            assert logs[0].type == LogType.Table
            assert logs[0].data == {"batch_size":4, "epochs":1000}
            ```

        Arguments:
            limit (int, optional): limit of logs to retrieve. Defaults to None.
            offset (int, optional): offset to start retrieving logs. Defaults to None.

        Returns:
            A list of (Log) objects
        """
        return TrainingService(self.connexion, self.experiment_id).list_logs(
            limit, offset
        )

    @exception_handler
    @beartype
    def get_log(self, name: str) -> Log:
        """Retrieve a (Log) from its name in this (FastTraining).

        Arguments:
            name (str): name of the log to retrieve
        Returns:
            A (Log) object
        """
        return TrainingService(self.connexion, self.experiment_id).get_log(name)

    @exception_handler
    @beartype
    def get_parameters(self) -> dict:
        """Retrieve parameters of this (FastTraining)

        Return:
            a dict
        """
        return TrainingService(self.connexion, self.experiment_id).get_parameters()

    @exception_handler
    @beartype
    def list_artifacts(
        self,
        limit: int | None = None,
        offset: int | None = None,
        order_by: list[str] | None = None,
    ) -> list[Artifact]:
        """List artifacts stored for this (FastTraining).

        Arguments:
            limit (int, optional): limit of artifacts to retrieve. Defaults to None.
            offset (int, optional): offset to start retrieving artifacts. Defaults to None.
            order_by (list[str], optional): fields to order by. Defaults to None.

        Returns:
            A list of (Artifact) objects that you can manipulate
        """
        return TrainingService(self.connexion, self.experiment_id).list_artifacts(
            limit, offset, order_by
        )



import logging
from abc import ABC, abstractmethod
from uuid import UUID

from beartype import beartype
from orjson import orjson

from picsellia.decorators import exception_handler
from picsellia.sdk.connexion import Connection
from picsellia.sdk.user import User

logger = logging.getLogger("picsellia")


class Resource(ABC):
    def __init__(self, resource_name: str):
        self.resource_name = resource_name

    @property
    @abstractmethod
    def id(self) -> UUID:
        pass

    @property
    @abstractmethod
    def connexion(self) -> Connection:
        pass

    @exception_handler
    @beartype
    def list_users(self) -> list[User]:
        """List all users of this resource

        Examples:
            ```python
            resource.list_users()
            ```

        Returns:
            list of (User) objects
        """
        r = self.connexion.get(
            f"/api/access/{self.resource_name}/{self.id}/users"
        ).json()
        return [User(self.connexion, item) for item in r["users"]]

    @exception_handler
    @beartype
    def add_user(self, user: User, role: str | None = None) -> None:
        """Add given user to this resource, with given role.
        If user is UNPRIVILEGED on Organization, then its role will be forced to LABELER.
        If role is None, user will be assigned configured role on Organization, or if not configured, USER.

        :warning: **DANGER ZONE**: This method will create access for a user. Role depends on type of resource you're adding the user.

        Examples:
            ```python
            user = client.find_user("foo")
            resource.add_user(user, "ADMIN")
            ```
        """
        payload = {"user_id": user.id}
        if role:
            payload["role"] = role

        self.connexion.post(
            f"/api/access/{self.resource_name}/{self.id}/users",
            data=orjson.dumps(payload),
        ).json()
        logger.info(f"User {user.username} has been added to this resource.")



import abc
from typing import Generic, TypeVar
from uuid import UUID

from picsellia.sdk.asset import Asset, MultiAsset
from picsellia.sdk.connexion import Connection
from picsellia.sdk.dao import Dao
from picsellia.sdk.data import Data, MultiData
from picsellia.sdk.multi_object import MultiObject
from picsellia.sdk.polygon import Polygon
from picsellia.sdk.rectangle import Rectangle
from picsellia.services.lister.asset_lister import AssetFilter, AssetLister
from picsellia.services.lister.data_lister import DataFilter, DataLister

TItem = TypeVar("TItem", bound=Dao)


class VisualSearchService(abc.ABC, Generic[TItem]):
    def __init__(self, connexion: Connection):
        self.connexion = connexion

    def query_visual_search(self, query: str, limit: int) -> MultiObject[TItem]:
        r = self.connexion.get(
            self.build_url_query(),
            params={
                "visual_search_query": query,
                "visual_search_limit": limit,
                # TODO: this field is already deprecated
                "visual_search_by": "similar",
            },
        ).json()
        result = {}
        data_scores = {}
        for point in r["points"]:
            data_scores[point["id"]] = point["score"]
            result[point["id"]] = None

        items = self.list_items_from_ids(list(data_scores.keys()))
        for item in items:
            item._score = data_scores[str(item.id)]
            result[str(item.id)] = item

        ordered_data = [item for item in result.values() if item]
        return self.build_multi_object(ordered_data)

    def list_embeddings(
        self,
        limit: int,
        with_vector: bool,
        with_payload: bool,
        has_error: bool | None,
    ) -> list[dict]:
        points = []
        offset = None
        while len(points) < limit:
            r = self.connexion.get(
                self.build_url_list(),
                params={
                    "with_vector": with_vector,
                    "with_payload": with_payload,
                    "has_error": has_error,
                    "limit": min(limit, 100),
                    "offset": offset,
                    **self.get_additional_scroll_params(),
                },
            ).json()
            points.extend(r["points"])
            offset = r["next_page_offset"]
            if offset is None:
                break

        return points

    @abc.abstractmethod
    def build_url_query(self) -> str:
        pass

    @abc.abstractmethod
    def build_url_list(self) -> str:
        pass

    @abc.abstractmethod
    def list_items_from_ids(self, ids: list[str]) -> list[Data]:
        pass

    @abc.abstractmethod
    def build_multi_object(self, items: list[TItem]) -> MultiObject[TItem]:
        pass

    @staticmethod
    def get_additional_scroll_params() -> dict:
        return {}


class DatalakeVisualSearchService(VisualSearchService[Data]):
    def __init__(self, connexion: Connection, datalake_id: UUID):
        super().__init__(connexion)
        self.datalake_id = datalake_id

    def build_url_query(self) -> str:
        return f"/api/datalake/{self.datalake_id}/datas/ids/visual-search"

    def build_url_list(self) -> str:
        return f"/api/visual-search/datalake/{self.datalake_id}/points/scroll"

    def list_items_from_ids(self, ids: list[str]) -> list[Data]:
        return DataLister(self.connexion, self.datalake_id).list_items(
            filters=DataFilter.model_validate({"ids": ids})
        )

    def build_multi_object(self, items: list[Data]) -> MultiData:
        return MultiData(self.connexion, datalake_id=self.datalake_id, items=items)


class DatasetVersionVisualSearchService(VisualSearchService[Asset]):
    def __init__(self, connexion: Connection, dataset_version_id: UUID):
        super().__init__(connexion)
        self.dataset_version_id = dataset_version_id

    def build_url_query(self) -> str:
        return f"/api/dataset/version/{self.dataset_version_id}/datas/ids/visual-search"

    def build_url_list(self) -> str:
        return f"/api/visual-search/dataset-version/{self.dataset_version_id}/points/scroll"

    def list_items_from_ids(self, ids: list[str]) -> list[Asset]:
        return AssetLister(self.connexion, self.dataset_version_id).list_items(
            filters=AssetFilter.model_validate({"ids": ids})
        )

    def build_multi_object(self, items: list[Asset]) -> MultiAsset:
        return MultiAsset(self.connexion, self.dataset_version_id, items=items)


class RectangleVisualSearchService(VisualSearchService[Rectangle]):
    def __init__(self, connexion: Connection, dataset_version_id: UUID):
        super().__init__(connexion)
        self.dataset_version_id = dataset_version_id

    def build_url_query(self) -> str:
        raise NotImplementedError()

    def list_items_from_ids(self, ids: list[str]) -> list[Rectangle]:
        raise NotImplementedError()

    def build_multi_object(self, items: list[Rectangle]) -> MultiObject[Rectangle]:
        raise NotImplementedError()

    def build_url_list(self) -> str:
        return f"/api/dataset/version/{self.dataset_version_id}/embeddings/shapes"

    @staticmethod
    def get_additional_scroll_params() -> dict:
        return {"shape_type": "rectangle"}


class PolygonVisualSearchService(VisualSearchService[Polygon]):
    def __init__(self, connexion: Connection, dataset_version_id: UUID):
        super().__init__(connexion)
        self.dataset_version_id = dataset_version_id

    def build_url_query(self) -> str:
        raise NotImplementedError()

    def list_items_from_ids(self, ids: list[str]) -> list[Polygon]:
        raise NotImplementedError()

    def build_multi_object(self, items: list[Polygon]) -> MultiObject[Polygon]:
        raise NotImplementedError()

    def build_url_list(self) -> str:
        return f"/api/dataset/version/{self.dataset_version_id}/embeddings/shapes"

    @staticmethod
    def get_additional_scroll_params() -> dict:
        return {"shape_type": "polygon"}



import logging
from uuid import UUID

import orjson
from beartype import beartype

from picsellia.colors import Colors
from picsellia.decorators import exception_handler
from picsellia.sdk.connexion import Connection
from picsellia.sdk.dao import Dao
from picsellia.sdk.label import Label
from picsellia.types.schemas import LineSchema

logger = logging.getLogger("picsellia")


class Line(Dao):
    def __init__(
        self,
        connexion: Connection,
        dataset_version_id: UUID,
        annotation_id: UUID,
        data: dict,
    ) -> None:
        self._dataset_version_id = dataset_version_id
        self._annotation_id = annotation_id
        Dao.__init__(self, connexion, data)

    @property
    def annotation_id(self) -> UUID:
        """UUID of (Annotation) holding this (Line)"""
        return self._annotation_id

    @property
    def coords(self) -> list[list[int]]:
        """Coords of this (Line)"""
        return self._coords

    @property
    def label(self) -> Label:
        """(Label) of this (Line)"""
        return self._label

    @property
    def text(self) -> str | None:
        """OCR text of this (Polygon)"""
        return self._text

    def __str__(self):
        return f"{Colors.BLUE}Line ({len(self.coords)} points) with label {self.label.name} on annotation {self.annotation_id} {Colors.ENDC} (id: {self.id})"

    @exception_handler
    @beartype
    def sync(self) -> dict:
        r = self.connexion.get(f"/api/line/{self.id}").json()
        self.refresh(r)
        return r

    @exception_handler
    @beartype
    def refresh(self, data: dict) -> LineSchema:
        schema = LineSchema(**data)
        self._coords = schema.coords
        self._text = schema.text
        self._label = Label(
            self.connexion, self._dataset_version_id, schema.label.model_dump()
        )
        return schema

    @exception_handler
    @beartype
    def update(
        self,
        coords: list | None = None,
        label: Label | None = None,
        text: str | None = None,
    ) -> None:
        """Update this line with new coords or new label.

        Examples:
            ```python
            line.update(coords=[[0, 0], [0, 1], [1, 1]])
            ```

        Arguments:
            coords (List, optional): New coords of this (Line). Defaults to None.
            label (Label, optional): New label of this (Line). Defaults to None.
            text (str, optional): New ocr text of this (Line). Defaults to None.
        """
        payload = {}
        if coords is not None:
            payload["line"] = coords
        if label is not None:
            payload["label_id"] = label.id
        if text is not None:
            payload["text"] = text
        assert payload != {}, "You can't update this line with no data to update"
        r = self.connexion.patch(
            f"/api/line/{self.id}", data=orjson.dumps(payload)
        ).json()
        self.refresh(r)
        logger.info(f"{self} updated")

    @exception_handler
    @beartype
    def delete(self) -> None:
        """Delete this line from the platform.

        :warning: **DANGER ZONE**: Be very careful here!

        Examples:
            ```python
            line.delete()
            ```
        """
        self.connexion.delete(f"/api/line/{self.id}")
        logger.info(f"{self} deleted.")



import logging

import orjson
from beartype import beartype

import picsellia.utils as utils
from picsellia.colors import Colors
from picsellia.decorators import exception_handler
from picsellia.sdk.connexion import Connection
from picsellia.sdk.dao import Dao
from picsellia.sdk.model_version import ModelVersion
from picsellia.sdk.tag import Tag, TagTarget
from picsellia.sdk.taggable import Taggable
from picsellia.types.enums import Framework, InferenceType
from picsellia.types.schemas import ModelSchema

logger = logging.getLogger("picsellia")


class Model(Dao, Taggable):
    def __init__(self, connexion: Connection, data: dict):
        Dao.__init__(self, connexion, data)
        Taggable.__init__(self, TagTarget.MODEL)

    @property
    def name(self) -> str:
        """Name of this (Model)"""
        return self._name

    @property
    def private(self) -> bool:
        """Privacy of this (Model)"""
        return self._private

    def __str__(self):
        is_private = "" if self._private else "[PUBLIC]"
        return f"{Colors.BLUE}{is_private} Model '{self.name}' {Colors.ENDC} (id: {self.id})"

    @exception_handler
    @beartype
    def get_resource_url_on_platform(self) -> str:
        """Get platform url of this resource.

        Examples:
            ```python
            print(model.get_resource_url_on_platform())
            >>> "https://app.picsellia.com/model/62cffb84-b92c-450c-bc37-8c4dd4d0f590"
            ```

        Returns:
            Url on Platform for this resource
        """

        return f"{self.connexion.host}/model/{self.id}"

    @exception_handler
    @beartype
    def sync(self) -> dict:
        r = self.connexion.get(f"/api/model/{self.id}").json()
        self.refresh(r)
        return r

    @exception_handler
    @beartype
    def refresh(self, data: dict):
        schema = ModelSchema(**data)
        self._name = schema.name
        self._private = schema.private
        return schema

    @exception_handler
    @beartype
    def update(
        self,
        name: str | None = None,
        framework: str | Framework | None = None,
        private: bool | None = None,
        description: str | None = None,
        type: str | InferenceType | None = None,
    ) -> None:
        """Update a model with a new name, framework, privacy, description or type

        Examples:
            ```python
            model.update(description="Very cool model")
            ```

        Arguments:
            name (str): New name of the model
            framework (str, Framework): New framework of the model
            private (bool): New privacy of the model
            description (str): New description of the model
            type (str, InferenceType): New type of the model
        """
        if type:
            logging.warning(
                "'type' parameter is deprecated and will be removed in future versions. "
                "If you want to give a type to your model version, call update() on ModelVersion."
            )

        if framework:
            logging.warning(
                "'framework' parameter is deprecated and will be removed in future versions. "
                "If you want to give a framework to your model version, call update() on your ModelVersion."
            )

        if private is not None:
            logging.warning(
                "'private' parameter is deprecated and will be removed in future versions. "
                "You cannot update privacy of a model anymore."
            )

        payload = {
            "name": name,
            "description": description,
        }
        filtered_payload = utils.filter_payload(payload)
        r = self.connexion.patch(
            f"/api/model/{self.id}", data=orjson.dumps(filtered_payload)
        ).json()
        self.refresh(r)
        logger.info(f"{self} updated.")

    @exception_handler
    @beartype
    def delete(self) -> None:
        """Delete model.

        Delete the model in Picsellia database

        Examples:
            ```python
            model.delete()
            ```
        """
        self.connexion.delete(f"/api/model/{self.id}")
        logger.info(f"{self} deleted.")

    @exception_handler
    @beartype
    def get_tags(self) -> list[Tag]:
        """Retrieve the tags of your model.

        Examples:
            ```python
            tags = my_model.get_tags()
            assert tags[0].name == "my-model-1"
            ```

        Returns:
            A list of (Tag) object
        """
        r = self.sync()
        return [Tag(self.connexion, item) for item in r["tags"]]

    @exception_handler
    @beartype
    def create_version(
        self,
        docker_image_name: str | None = None,
        docker_flags: list[str] | None = None,
        thumb_object_name: str | None = None,
        notebook_link: str | None = None,
        labels: dict | None = None,
        base_parameters: dict | None = None,
        docker_env_variables: dict | None = None,
        name: str | None = None,
        framework: str | Framework | None = None,
        type: str | InferenceType | None = None,
        description: str | None = None,
        docker_tag: str | None = None,
    ) -> ModelVersion:
        """Create a version of a model.

        The version number of this model will be defined by the platform. It is incremented automatically.

        Examples:
            ```python
            model_v0 = model.create_version(labels={"1": "cat", "2": "dog"}, framework=Framework.TENSORFLOW, type=InferenceType.OBJECT_DETECTION)
            ```

        Arguments:
            docker_image_name (str, optional): Docker image name of this version. Defaults to None.
            docker_flags (list[str], optional): Docker flags of this version. Defaults to None.
            thumb_object_name (str, optional): Thumbnail object name of this version. Defaults to None.
            notebook_link (str, optional): Notebook link of this version. Defaults to None.
            labels (dict, optional): Labels of this version. Defaults to None.
            base_parameters (dict, optional): Base parameters of this version. Defaults to None.
            docker_env_variables (dict, optional): Docker environment variables of this version. Defaults to None.
            name (str, optional): Name of this version. Defaults to None.
            framework (str, Framework, optional): Framework of this version (tensorflow, pytorch, etc.). Defaults to None.
            type (str, InferenceType, optional): Type of this version (classification, object_detection, segmentation). Defaults to None.
            description (str, optional): Description of this version. Defaults to None.
            docker_tag (str, optional): Docker tag of this version. Defaults to None.

        Returns:
            A (ModelVersion) object
        """
        if framework:
            framework = Framework.validate(framework)

        if type:
            type = InferenceType.validate(type)

        if type and type not in [
            InferenceType.CLASSIFICATION,
            InferenceType.OBJECT_DETECTION,
            InferenceType.SEGMENTATION,
            InferenceType.LINE,
            InferenceType.POINT,
            InferenceType.KEYPOINT
        ]:
            raise TypeError(f"Type '{type}' not supported yet for model version")

        payload = {
            "name": name,
            "type": type,
            "framework": framework,
            "docker_image": docker_image_name,
            "docker_flags": docker_flags,
            "thumbnail_object_name": thumb_object_name,
            "notebook_link": notebook_link,
            "labels": labels,
            "parameters": base_parameters,
            "docker_env_variables": docker_env_variables,
            "description": description,
            "docker_tag": docker_tag,
        }
        filtered_payload = utils.filter_payload(payload)
        r = self.connexion.post(
            f"/api/model/{self.id}/versions",
            data=orjson.dumps(filtered_payload),
        ).json()
        return ModelVersion(self.connexion, r)

    @exception_handler
    @beartype
    def get_version(self, version: int | str) -> ModelVersion:
        """Retrieve a version of a model from its version or its name

        Examples:
            ```python
            # Assuming model is a Model without version
            model_version_a = model.create_version("first-version")
            model_version_b = model.get_version(0)
            model_version_c = model.get_version("first-version")
            assert model_version_a == model_version_b
            assert model_version_a == model_version_c
            ```

        Arguments:
            version (int, str): Version number or name of the version

        Returns:
            A (ModelVersion) object
        """
        if isinstance(version, str):
            params = {"name": version}
        else:
            params = {"version": version}
        r = self.connexion.get(
            f"/api/model/{self.id}/versions/find", params=params
        ).json()
        return ModelVersion(self.connexion, r)

    @exception_handler
    @beartype
    def list_versions(
        self,
        limit: int | None = None,
        offset: int | None = None,
        order_by: list[str] | None = None,
    ) -> list[ModelVersion]:
        """List versions of this model.

        Examples:
            ```python
            versions = model.list_versions()
            ```

        Arguments:
            limit (int, optional): Limit of versions to retrieve. Defaults to None.
            offset (int, optional): Offset to start retrieving versions. Defaults to None.
            order_by (list[str], optional): fields to order by. Defaults to None.

        Returns:
            A list of (ModelVersion) object of this model
        """
        params = {"limit": limit, "offset": offset, "order_by": order_by}
        params = utils.filter_payload(params)
        r = self.connexion.get(f"/api/model/{self.id}/versions", params=params).json()
        return [ModelVersion(self.connexion, item) for item in r["items"]]



import logging

import orjson
from beartype import beartype
from deprecation import deprecated

from picsellia.colors import Colors
from picsellia.decorators import exception_handler
from picsellia.sdk.connexion import Connection
from picsellia.sdk.dao import Dao
from picsellia.types.enums import (
    Framework,
    InferenceType,
    ProcessingInputType,
    ProcessingType,
)
from picsellia.types.schemas import ProcessingSchema
from picsellia.utils import filter_payload

logger = logging.getLogger("picsellia")


class Processing(Dao):
    def __init__(self, connexion: Connection, data: dict):
        Dao.__init__(self, connexion, data)

    @property
    def name(self) -> str:
        """Name of this (Processing)"""
        return self._name

    @property
    def type(self) -> ProcessingType:
        """Type of this (Processing)"""
        return self._type

    @property
    def docker(self) -> str:
        """Docker image of this (Processing)"""
        return f"{self._docker_image}:{self._docker_tag}"

    def __str__(self):
        return f"{Colors.GREEN}Processing '{self.name}' {Colors.ENDC} (id: {self.id})"

    @exception_handler
    @beartype
    def sync(self) -> dict:
        r = self.connexion.get(f"/api/processing/{self.id}").json()
        self.refresh(r)
        return r

    @exception_handler
    @beartype
    def refresh(self, data: dict) -> ProcessingSchema:
        schema = ProcessingSchema(**data)
        self._name = schema.name
        self._type = schema.type
        self._docker_image = schema.docker_image
        self._docker_tag = schema.docker_tag
        return schema

    @exception_handler
    @beartype
    def update(
        self,
        docker_image: str | None = None,
        docker_tag: str | None = None,
        description: str | None = None,
        default_parameters: dict | None = None,
        default_cpu: int | None = None,
        default_gpu: int | None = None,
    ) -> None:
        """Update docker_image, description or default_parameters of (Processing).

        Examples:
            ```python
            processing.update(docker_image='new-image', docker_tag='1.2.0')
            ```

        Arguments:
            docker_image (str, optional): New docker image of this (Processing). Defaults to None.
            docker_tag (str, optional): New docker tag of this (Processing). Defaults to None.
            description (str, optional): New description of the (Processing). Defaults to None.
            default_parameters (dict, optional): New default parameters of the (Processing). Defaults to None.
            default_cpu (str or InferenceType, optional): New default cpu of the (Processing). Defaults to None.
            default_gpu (str or InferenceType, optional): New default gpu of the (Processing). Defaults to None.
        """
        payload = {
            "docker_image": docker_image,
            "docker_tag": docker_tag,
            "description": description,
            "default_parameters": default_parameters,
            "default_cpu": default_cpu,
            "default_gpu": default_gpu,
        }
        self._update_processing(payload)
        logger.info(f"{self} updated")

    def _update_processing(self, payload: dict) -> None:
        filtered_payload = filter_payload(payload)
        r = self.connexion.patch(
            f"/api/processing/{self.id}", data=orjson.dumps(filtered_payload)
        ).json()
        self.refresh(r)

    @exception_handler
    @beartype
    def delete(self) -> None:
        """Delete this processing from the platform.

        :warning: **DANGER ZONE**: Be very careful here!

        Examples:
            ```python
            processing.delete()
            ```
        """
        self.connexion.delete(f"/api/processing/{self.id}")
        logger.info(f"{self} deleted.")

    @exception_handler
    @beartype
    def list_processing_inputs(self) -> list[dict]:
        """
        List inputs of this processing.
        """
        r = self.connexion.get(f"/api/processing/{self.id}/inputs")
        return r.json()["items"]

    @exception_handler
    @beartype
    def add_processing_input(
        self,
        name: str,
        input_type: ProcessingInputType,
        required: bool,
        inference_type_constraint: InferenceType | None,
        framework_constraint: Framework | None,
    ) -> dict:
        """
        Add an input to this processing. Inputs are checked on launch_processing(), they must match types and constraints.
        They are passed to the context of your processing in "inputs" key.
        """
        payload = {
            "name": name,
            "input_type": input_type,
            "required": required,
            "inference_type_constraint": inference_type_constraint,
            "framework_constraint": framework_constraint,
        }
        r = self.connexion.post(
            f"/api/processing/{self.id}/inputs", data=orjson.dumps(payload)
        )
        return r.json()

    @exception_handler
    @beartype
    def update_processing_input(
        self,
        name: str,
        required: bool,
        inference_type_constraint: InferenceType | None,
        framework_constraint: Framework | None,
    ) -> dict:
        """
        Update an input of this processing.
        """
        payload = {
            "required": required,
            "inference_type_constraint": inference_type_constraint,
            "framework_constraint": framework_constraint,
        }
        r = self.connexion.put(
            f"/api/processing/{self.id}/inputs/{name}", data=orjson.dumps(payload)
        )
        return r.json()

    @exception_handler
    @beartype
    def delete_processing_input(self, name: str) -> None:
        """
        Delete an input from this processing.
        """
        self.connexion.delete(f"/api/processing/{self.id}/inputs/{name}")
        logger.info(f"Processing input {name} of {self} have been removed.")

    @exception_handler
    @beartype
    @deprecated(
        deprecated_in="6.29.0",
        details="In the new processing system, constraints are defined on Processing inputs. This method will update processing input that matches old system, but should not be used anymore.",
    )
    def create_dataset_version_processing_constraints(
        self,
        input_dataset_version_type: str | InferenceType | None = None,
        output_dataset_version_type: str | InferenceType | None = None,
        model_version_framework: str | Framework | None = None,
        model_version_type: str | InferenceType | None = None,
    ):
        """
        This method can only be used for processing running on DatasetVersion.
        So the type of this Processing must be one of:
            - PRE_ANNOTATION
            - DATA_AUGMENTATION
            - DATASET_VERSION_CREATION
            - AUTO_TAGGING
            - AUTO_ANNOTATION
        """

        # this is a blacklist to be future compatible
        if self.type in [
            ProcessingType.DATA_AUTO_TAGGING,
            ProcessingType.MODEL_COMPRESSION,
            ProcessingType.MODEL_CONVERSION,
        ]:
            raise TypeError(
                f"{self.type} processings cannot have this type of constraints"
            )

        payload = {
            "dataset_version_constraints": {
                "input_dataset_version_type": InferenceType.validate(
                    input_dataset_version_type
                )
                if input_dataset_version_type
                else None,
                "output_dataset_version_type": InferenceType.validate(
                    output_dataset_version_type
                )
                if output_dataset_version_type
                else None,
                "model_version_framework": Framework.validate(model_version_framework)
                if model_version_framework
                else None,
                "model_version_type": InferenceType.validate(model_version_type)
                if model_version_type
                else None,
            }
        }
        self._update_processing(payload)
        logger.info(f"Constraints added on {self}")

    @exception_handler
    @beartype
    @deprecated(
        deprecated_in="6.29.0",
        details="In the new processing system, constraints are defined on Processing inputs. This method will update processing input that matches old system, but should not be used anymore.",
    )
    def create_datalake_processing_constraints(
        self,
        model_version_framework: str | Framework | None = None,
        model_version_type: str | InferenceType | None = None,
    ):
        """
        This method can only be used for processing running on Datalake.
        So the type of this Processing must be one of:
            - DATA_AUTO_TAGGING
        """
        # this is a blacklist to be future compatible
        if self.type in [
            ProcessingType.PRE_ANNOTATION,
            ProcessingType.DATA_AUGMENTATION,
            ProcessingType.DATASET_VERSION_CREATION,
            ProcessingType.AUTO_TAGGING,
            ProcessingType.AUTO_ANNOTATION,
            ProcessingType.MODEL_COMPRESSION,
            ProcessingType.MODEL_CONVERSION,
        ]:
            raise TypeError(
                f"{self.type} processings cannot have this type of constraints"
            )

        payload = {
            "datalake_constraints": {
                "model_version_framework": Framework.validate(model_version_framework)
                if model_version_framework
                else None,
                "model_version_type": InferenceType.validate(model_version_type)
                if model_version_type
                else None,
            }
        }
        self._update_processing(payload)
        logger.info(f"Constraints added on {self}")



import logging
import os
from pathlib import Path
from typing import Any
from uuid import UUID

import orjson
from beartype import beartype
from deprecation import deprecated

from picsellia import exceptions, utils
from picsellia.colors import Colors
from picsellia.decorators import exception_handler
from picsellia.exceptions import (
    BadRequestError,
    ResourceConflictError,
    ResourceNotFoundError,
)
from picsellia.sdk.artifact import Artifact
from picsellia.sdk.asset import Asset, MultiAsset
from picsellia.sdk.connexion import Connection
from picsellia.sdk.dao import Dao
from picsellia.sdk.dataset_version import DatasetVersion
from picsellia.sdk.evaluation import Evaluation, MultiEvaluation
from picsellia.sdk.job import Job
from picsellia.sdk.label import Label
from picsellia.sdk.log import Log, LogType
from picsellia.sdk.logging_file import LoggingFile
from picsellia.sdk.model import Model
from picsellia.sdk.model_version import ModelVersion
from picsellia.sdk.worker import Worker
from picsellia.services.training import TrainingService
from picsellia.types.enums import (
    AddEvaluationType,
    AnnotationStatus,
    ExperimentStatus,
    InferenceType,
    JobStatus,
    ObjectDataType,
)
from picsellia.types.schemas import ExperimentSchema, LogDataType

logger = logging.getLogger("picsellia")


class Experiment(Dao):
    def __init__(self, connexion: Connection, data: dict) -> None:
        Dao.__init__(self, connexion, data)

    @property
    def name(self) -> str:
        """Name of this (Experiment)"""
        return self._name

    @property
    def project_id(self) -> UUID:
        """Project id of this (Experiment)"""
        return self._project_id

    @property
    def status(self) -> ExperimentStatus:
        """Status of this (Experiment)"""
        return self._status

    @property
    def base_dir(self):
        return self.name

    @property
    def metrics_dir(self):
        return os.path.join(self.base_dir, "metrics")

    @property
    def png_dir(self):
        return os.path.join(self.base_dir, "images")

    @property
    def checkpoint_dir(self):
        return os.path.join(self.base_dir, "checkpoint")

    @property
    def record_dir(self):
        return os.path.join(self.base_dir, "records")

    @property
    def config_dir(self):
        return os.path.join(self.base_dir, "config")

    @property
    def results_dir(self):
        return os.path.join(self.base_dir, "results")

    @property
    def exported_model_dir(self):
        return os.path.join(self.base_dir, "exported_model")

    def __str__(self):
        return f"{Colors.BLUE}Experiment '{self.name}' {Colors.ENDC} (id: {self.id})"

    @exception_handler
    @beartype
    def get_resource_url_on_platform(self) -> str:
        """Get platform url of this resource.

        Examples:
            ```python
            print(experiment.get_resource_url_on_platform())
            >>> "https://app.picsellia.com/experiment/62cffb84-b92c-450c-bc37-8c4dd4d0f590"
            ```

        Returns:
            Url on Platform for this resource
        """

        return f"{self.connexion.host}/experiment/{self.id}"

    @exception_handler
    @beartype
    def refresh(self, data: dict):
        schema = ExperimentSchema(**data)
        self._name = schema.name
        self._project_id = schema.project_id
        self._status = schema.status
        return schema

    @exception_handler
    @beartype
    def sync(self) -> dict:
        r = self.connexion.get(f"/api/experiment/{self.id}").json()
        self.refresh(r)
        return r

    @exception_handler
    @beartype
    def update(
        self,
        name: str | None = None,
        description: str | None = None,
        base_experiment: "Experiment | None" = None,
        base_model_version: "ModelVersion | None" = None,
        status: ExperimentStatus | str | None = None,
    ) -> None:
        """Update this experiment with a given name, description, a base experiment, a base model version or a status.

        Examples:
            ```python
            my_experiment.update(description="First try Yolov5")
            ```

        Arguments:
            name (str, optional): Name of the experiment. Defaults to None.
            description (str, optional): Description of the experiment. Defaults to None.
            base_experiment (Experiment, optional): deprecated.
            base_model_version (ModelVersion, optional): deprecated.
            status (ExperimentStatus or str, optional): Status of the experiment. Defaults to None.
        """
        payload: dict[str, Any] = {"name": name, "description": description}

        if status:
            payload["status"] = ExperimentStatus.validate(status)

        if base_experiment:
            logger.warning(
                "base_experiment parameter is deprecated and cannot be used anymore. Please call attach_experiment()"
            )

        if base_model_version:
            logger.warning(
                "base_model_version parameter is deprecated and cannot be used anymore. Please call attach_model_version()"
            )

        filtered_payload = utils.filter_payload(payload)
        r = self.connexion.patch(
            f"/api/experiment/{self.id}", data=orjson.dumps(filtered_payload)
        ).json()
        self.refresh(r)

    @exception_handler
    @beartype
    def delete(self) -> None:
        """Delete this experiment.

        Examples:
            ```python
            my_experiment.delete()
            ```
        """
        self.connexion.delete(f"/api/experiment/{self.id}")

    @exception_handler
    @beartype
    def list_artifacts(
        self,
        limit: int | None = None,
        offset: int | None = None,
        order_by: list[str] | None = None,
    ) -> list[Artifact]:
        """List artifacts stored in the experiment.

        Examples:
            ```python
            artifacts = my_experiment.list_artifacts()
            ```

        Arguments:
            limit (int, optional): limit of artifacts to retrieve. Defaults to None.
            offset (int, optional): offset to start retrieving artifacts. Defaults to None.
            order_by (list[str], optional): fields to order by. Defaults to None.

        Returns:
            A list of (Artifact) objects that you can manipulate
        """
        return TrainingService(self.connexion, self.id).list_artifacts(
            limit, offset, order_by
        )

    @exception_handler
    @beartype
    def delete_all_artifacts(self) -> None:
        """Delete all stored artifacts for experiment

        :warning: **DANGER ZONE**: This will definitely remove the artifacts from our servers

        Examples:
            ```python
            experiment.delete_all_artifacts()
            ```
        """
        artifact_ids = self.connexion.get(
            f"/api/experiment/{self.id}/artifacts/ids"
        ).json()
        payload = {"ids": artifact_ids}
        self.connexion.delete(
            f"/api/experiment/{self.id}/artifacts", data=orjson.dumps(payload)
        )

    @exception_handler
    @beartype
    def create_artifact(
        self, name: str, filename: str, object_name: str, large: bool = False
    ) -> Artifact:
        """Create an artifact for this experiment.

        Examples:
            ```python
            self.create_artifact(name="a_file", filename="file.png", object_name="some_file_in_s3.png", large=False)
            ```
        Arguments:
            name (str): name of the artifact.
            filename (str): filename.
            object_name (str): s3 object name.
            large (bool, optional): >5Mb or not. Defaults to False.

        Returns:
            An (Artifact) object
        """
        payload = {
            "name": name,
            "filename": filename,
            "object_name": object_name,
            "large": large,
        }
        r = self.connexion.post(
            f"/api/experiment/{self.id}/artifacts", data=orjson.dumps(payload)
        ).json()
        return Artifact(self.connexion, r)

    @exception_handler
    @beartype
    def _create_or_update_file(
        self, name: str, filename: str, object_name: str, large: bool
    ) -> Artifact:
        try:
            stored = self.get_artifact(name)
            stored.update(
                name=name, filename=filename, object_name=object_name, large=large
            )
            return stored
        except exceptions.ResourceNotFoundError:
            return self.create_artifact(
                name=name, filename=filename, object_name=object_name, large=large
            )

    @exception_handler
    @beartype
    def store(
        self, name: str, path: str | Path | None = None, do_zip: bool = False
    ) -> Artifact:
        """Store an artifact and attach it to the experiment.

        Examples:
            ```python
            # Zip and store a folder as an artifact for the experiment
            # you can choose an arbitrary name or refer to our 'namespace'
            # for certain artifacts to have a custom behavior

            trained_model_path = "my_experiment/saved_model"
            experiment.store("model-latest", trained_model_path, do_zip=True)
            ```
        Arguments:
            name (str): name for the artifact. Defaults to "".
            path (str or Path): path to the file or folder. Defaults to None.
            do_zip (bool, optional): Whether to compress the file to a zip file. Defaults to False.

        Raises:
            FileNotFoundException: No file found at the given path

        Returns:
            An (Artifact) object
        """
        if path is None:  # pragma: no cover
            return self.store_local_artifact(name)

        if do_zip:
            path = utils.zip_dir(path)

        filename = os.path.basename(path)
        object_name = self.connexion.generate_experiment_object_name(
            filename, ObjectDataType.ARTIFACT, self.id
        )
        _, is_large, _ = self.connexion.upload_file(object_name, path)

        return self._create_or_update_file(
            name=name, filename=filename, object_name=object_name, large=is_large
        )

    @exception_handler
    @beartype
    def store_local_artifact(self, name: str) -> Artifact:  # pragma: no cover
        """Store an artifact in platform that is locally stored

        This artifact shall have the name: config, checkpoint-data-latest, checkpoint-index-latest or model-latest

        It will look for special file into current directory.

        Examples:
            ```python
            my_experiment.store_local_artifact("model-latest")
            ```
        Arguments:
            name (str): Name of the artifact to store

        Returns:
            An (Artifact) object
        """
        assert (
            name == "config"
            or name == "checkpoint-data-latest"
            or name == "checkpoint-index-latest"
            or name == "model-latest"
        ), "This name cannot be used to store an artifact"

        if name == "config":
            filename = "pipeline.config"
            path = os.path.join(self.config_dir, filename)
            if not os.path.isfile(path):
                raise FileNotFoundError(f"{path} not found")

        elif name == "checkpoint-data-latest":
            file_list = os.listdir(self.checkpoint_dir)
            ckpt_id = max(
                [int(p.split("-")[1].split(".")[0]) for p in file_list if "index" in p]
            )
            filename = None
            for f in file_list:
                if f"{ckpt_id}.data" in f:
                    filename = f
                    break
            if filename is None:
                raise exceptions.ResourceNotFoundError(
                    "Could not find matching data file with index"
                )
            path = os.path.join(self.checkpoint_dir, filename)

        elif name == "checkpoint-index-latest":
            file_list = os.listdir(self.checkpoint_dir)
            ckpt_id = max(
                [int(p.split("-")[1].split(".")[0]) for p in file_list if "index" in p]
            )
            filename = f"ckpt-{ckpt_id}.index"
            path = os.path.join(self.checkpoint_dir, filename)

        elif name == "model-latest":  # pragma: no cover
            file_path = os.path.join(self.exported_model_dir, "saved_model")
            path = utils.zip_dir(file_path)
            filename = "saved_model.zip"

        else:
            raise RuntimeError("unreachable code")

        object_name = self.connexion.generate_experiment_object_name(
            filename, ObjectDataType.ARTIFACT, self.id
        )
        _, is_large, _ = self.connexion.upload_file(object_name, path)

        return self._create_or_update_file(
            name=name, filename=filename, object_name=object_name, large=is_large
        )

    @exception_handler
    @beartype
    def get_base_model_version(self) -> ModelVersion:
        """Retrieve the base model version of this experiment.

        Examples:
            ```python
            model_version = experiment.get_base_model_version()
            ```

        Returns:
            A (ModelVersion) object representing the base model.
        """
        r = self.sync()
        if r["base_model_version_id"] is None:
            raise exceptions.NoBaseModelVersionError(
                "There is no base model for this experiment."
            )
        r = self.connexion.get(
            f"/api/model/version/{r['base_model_version_id']}"
        ).json()
        return ModelVersion(self.connexion, r)

    @exception_handler
    @beartype
    def get_base_experiment(self) -> "Experiment":
        """Retrieve the base experiment of this experiment.

        Examples:
            ```python
            previous = experiment.get_base_experiment()
            ```

        Returns:
            An (Experiment) object representing the base experiment.
        """
        base_experiment_id = self.sync()["base_experiment_id"]
        if not base_experiment_id:
            raise exceptions.NoBaseExperimentError(
                "There is no base experiment for this experiment"
            )
        r = self.connexion.get(f"/api/experiment/{base_experiment_id}").json()
        return Experiment(self.connexion, r)

    @exception_handler
    @beartype
    def get_artifact(self, name: str) -> Artifact:
        """Retrieve an artifact from its name in this experiment.

        Examples:
            ```python
            model_artifact = experiment.get_artifact("model-latest")
            assert model_artifact.name == "model-latest"
            assert model_artifact.object_name == "d67924a0-7757-48ed-bf7a-322b745e917e/saved_model.zip"
            ```
        Arguments:
            name (str): Name of the artifact to retrieve

        Returns:
            An (Artifact) object
        """
        params = {"name": name}
        r = self.connexion.get(
            f"/api/experiment/{self.id}/artifacts/find", params=params
        ).json()
        return Artifact(self.connexion, r)

    @exception_handler
    @beartype
    def list_logs(
        self,
        limit: int | None = None,
        offset: int | None = None,
        order_by: list[str] | None = None,
    ) -> list:
        """List everything that has been logged.

        List everything that has been logged to an experiment.

        Examples:
            ```python
            logs = experiment.list_logs()
            assert logs[0].type == LogType.Table
            assert logs[0].data == {"batch_size":4, "epochs":1000}
            ```

        Arguments:
            limit (int, optional): limit of logs to retrieve. Defaults to None.
            offset (int, optional): offset to start retrieving logs. Defaults to None.
            order_by (list[str], optional): fields to order by. Defaults to None.

        Returns:
            A list of (Log) objects
        """
        return TrainingService(self.connexion, self.id).list_logs(
            limit, offset, order_by
        )

    @exception_handler
    @beartype
    def delete_all_logs(self) -> None:
        """Delete everything that has been logged.

        Delete everything that has been logged into this experiment.

        Examples:
            ```python
            experiment.delete_all_logs()
            ```
        """
        log_ids = self.connexion.get(f"/api/experiment/{self.id}/logs/ids").json()
        payload = {"ids": log_ids}
        self.connexion.delete(
            f"/api/experiment/{self.id}/logs", data=orjson.dumps(payload)
        )

    @exception_handler
    @beartype
    def get_log(self, name: str) -> Log:
        """Retrieve a log from its name in this experiment.

        Examples:
            ```python
            parameters = experiment.get_log("parameters")
            assert log.parameters == { "batch_size":4, "epochs":1000}
            ```
        Arguments:
            name (str): name of the log to retrieve

        Returns:
            A (Log) object
        """
        return TrainingService(self.connexion, self.id).get_log(name)

    @exception_handler
    @beartype
    def log(
        self,
        name: str,
        data: LogDataType,
        type: LogType | str,
        replace: bool = True,
    ) -> Log:
        """Record some data in an experiment.

        It will create a (Log) object, that you can manipulate in SDK.
        All logs are displayed in experiment view on Picsellia.

        If a (Log) with this name already exists, it will be updated unless parameter replace is set to False.

        If it is a LogType LINE and replace is True, then it will append data at the end of stored data.
        If you want to replace a line, delete this log and create another one.


        Examples:
            ```python
            parameters = {
                "batch_size":4,
                "epochs":1000
            }
            exp.log("parameters", parameters, type=LogType.TABLE)
            ```
        Arguments:
            name (str): Name of the log.
            data (Any): Data to be saved.
            type (LogType, optional): Type of the data to log.
                                  This will condition how it is displayed in the experiment dashboard. Defaults to None.
            replace (bool, optional): If true and log already exists and it is not a line, it will replace log data.
                                      Defaults to True.

        Raises:
            Exception: Impossible to upload the file when logging an image.

        Returns:
            A (Log) object
        """
        log_type = LogType.validate(type)

        if log_type == LogType.IMAGE:
            if not replace:
                # Assert log does not exist before pushing image into S3
                try:
                    self.get_log(name=name)
                    raise ResourceConflictError(
                        message="A log image with this name already exists. "
                        "If you want to replace it please give parameter replace=False"
                    )
                except ResourceNotFoundError:
                    pass

            if not isinstance(data, str):
                raise BadRequestError(
                    message="For log image, you need to give a path as data"
                )

            path = data
            filename = os.path.basename(path)
            object_name = self.connexion.generate_experiment_object_name(
                filename, ObjectDataType.LOG_IMAGE, self.id
            )
            _, large, _ = self.connexion.upload_file(object_name, path)
            data = {
                "object_name": object_name,
                "large": large,
                "filename": filename,
                "name": filename,
            }

        params = {"raise_on_conflict": (not replace)}
        payload = {"name": name, "data": data, "type": log_type.name}
        r = self.connexion.post(
            f"/api/experiment/{self.id}/logs", params=params, data=orjson.dumps(payload)
        ).json()
        return Log(self.connexion, r)

    @exception_handler
    @beartype
    def log_parameters(self, parameters: dict) -> Log:
        """Record parameters of an experiment into Picsellia.

        If parameters were already setup, it will be replaced.

        Examples:
            ```python
            parameters = {
                "batch_size":4,
                "epochs":1000
            }
            exp.log_parameters(parameters)
            ```
        Arguments:
            parameters (Any): Parameters to be saved.

        Returns:
            A (Log) object
        """
        return self.log("parameters", parameters, type=LogType.TABLE)

    @exception_handler
    @beartype
    def store_logging_file(self, path: str | Path) -> LoggingFile:
        """Store a logging file for this experiment.

        Examples:
            ```python
            experiment.store_logging_file("logs.txt")
            ```
        Arguments:
            path (str or Path): path to the file or folder.

        Raises:
            FileNotFoundException: No file found at the given path

        Returns:
            A (LoggingFile) object
        """
        if not os.path.exists(path):
            raise exceptions.FileNotFoundException(f"{path} not found")

        filename = os.path.basename(path)
        object_name = self.connexion.generate_experiment_object_name(
            filename, ObjectDataType.LOGGING, self.id
        )
        _, is_large, _ = self.connexion.upload_file(object_name, path)

        payload = {"object_name": object_name}

        r = self.connexion.post(
            f"/api/experiment/{self.id}/logging/save",
            data=orjson.dumps(payload),
        ).json()
        return LoggingFile(self.connexion, r)

    @exception_handler
    @beartype
    def get_logging_file(self) -> LoggingFile:
        """Retrieve logging file of this experiment.

        Examples:
            ```python
            logging_file = experiment.get_logging_file()
            logging_file.download()
            ```

        Returns:
            A (LoggingFile) object
        """
        r = self.connexion.get(f"/api/experiment/{self.id}/logging").json()
        return LoggingFile(self.connexion, r)

    @exception_handler
    @beartype
    def send_logging(
        self,
        log: str | list,
        part: str,
        final: bool = False,
        special: str | bool | list = False,
    ) -> None:
        """Send a logging experiment to the experiment.

        Examples:
            ```python
            experiment.send_logging("Hello World", "Hello", final=True)
            ```

        Arguments:
            log (str): Log content
            part (str): Logging Part
            final (bool, optional): True if Final line. Defaults to False.
            special (bool, optional): True if special log. Defaults to False.
        """
        if not hasattr(self, "line_nb"):
            self.line_nb = 0

        to_send = {
            "line_nb": self.line_nb,
            "log": log,
            "final": final,
            "part": part,
            "special": special,
        }
        self.line_nb += 1
        self.connexion.post(
            f"/api/experiment/{self.id}/logging",
            data=orjson.dumps(to_send),
        )

    @exception_handler
    @beartype
    def start_logging_chapter(self, name: str) -> None:
        """Print a log entry to the log.

        Examples:
            ```python
            experiment.start_logging_chapter("Training")
            ```

        Arguments:
            name (str): Chapter name
        """
        utils.print_start_chapter_name(name)
        utils.print_line_return()

    @exception_handler
    @beartype
    def start_logging_buffer(self, length: int = 1) -> None:
        """Start logging buffer.

        Examples:
            ```python
            experiment.start_logging_buffer()
            ```

        Arguments:
            length (int, optional): Buffer length. Defaults to 1.
        """
        utils.print_logging_buffer(length)
        self.buffer_length = length

    @exception_handler
    @beartype
    def end_logging_buffer(self) -> None:
        """End the logging buffer.

        Examples:
            ```python
            experiment.end_logging_buffer()
            ```
        """
        utils.print_logging_buffer(self.buffer_length)

    @exception_handler
    @beartype
    def update_job_status(self, status: JobStatus | str) -> None:
        """Update the job status of this experiment.

        Examples:
            ```python
            experiment.update_job_status(JobStatus.FAILED)
            ```

        Arguments:
            status (JobStatus): Status to send
        """
        to_send = {
            "status": JobStatus.validate(status),
        }
        self.connexion.patch(
            f"/api/experiment/{self.id}/job/status",
            data=orjson.dumps(to_send),
        )

    @exception_handler
    @beartype
    def export_as_model(self, name: str) -> ModelVersion:
        """Publish an Experiment as a ModelVersion to your registry.
           A Model with given name will be created, and its first version will be the exported experiment

        Examples:
            ```python
            model_version = experiment.export_as_model("awesome-model")
            ```
        Arguments:
            name (str): Target Name for the model in the registry.

        Returns:
            A (ModelVersion) just created from the experiment
        """
        payload = {"name": name}
        r = self.connexion.post(
            f"/api/experiment/{self.id}/publish", data=orjson.dumps(payload)
        ).json()
        model_version = ModelVersion(self.connexion, r)
        logger.info(f"Experiment published as {model_version}")
        return model_version

    @exception_handler
    @beartype
    def export_in_existing_model(self, existing_model: Model) -> ModelVersion:
        """Publish an Experiment as a (ModelVersion) of given already existing (Model)

        Examples:
            ```python
            my_model = client.get_model("foo_model")
            model_version = experiment.export_in_existing_model(my_model)
            ```
        Arguments:
            existing_model ((Model)): Model in the registry were this experiment should be exported.

        Returns:
            A (ModelVersion) just created from the experiment
        """
        payload = {"model_id": existing_model.id}
        r = self.connexion.post(
            f"/api/experiment/{self.id}/publish/version", data=orjson.dumps(payload)
        ).json()
        model_version = ModelVersion(self.connexion, r)
        logger.info(f"Experiment published as {model_version} in {existing_model}")
        return model_version

    @exception_handler
    @beartype
    def launch(self, gpus: int = 1) -> Job:
        """Launch a job on a remote environment with this experiment.

        :information-source: The remote environment has to be setup prior launching the experiment.
        It defaults to our remote training engine.

        Examples:
            ```python
            experiment.launch()
            ```
        Arguments:
            gpus (int, optional): Number of GPU to use for the training. Defaults to 1.
        """
        payload = {
            "gpus": gpus,
        }

        r = self.connexion.post(
            f"/api/experiment/{self.id}/launch", data=orjson.dumps(payload)
        ).json()
        logger.info("Job launched successfully")
        return Job(self.connexion, r, 2)

    def _setup_dirs(self):
        """Create the directories for the project."""
        if not os.path.isdir(self.name):
            logger.debug(
                "No directory for this project has been found, creating directory and sub-directories..."
            )
            os.mkdir(self.name)

        self._create_dir(self.base_dir)
        self._create_dir(self.png_dir)
        self._create_dir(self.checkpoint_dir)
        self._create_dir(self.metrics_dir)
        self._create_dir(self.record_dir)
        self._create_dir(self.config_dir)
        self._create_dir(self.results_dir)
        self._create_dir(self.exported_model_dir)

    @exception_handler
    @beartype
    def _create_dir(self, dir_name: str) -> None:
        """Create a directory if it does not exist.

        Arguments:
            dir_name (str): [directory name]
        """
        if not os.path.isdir(dir_name):
            os.mkdir(dir_name)

    @exception_handler
    @beartype
    def download_artifacts(self, with_tree: bool):
        """
        Download all artifacts from the experiment to the local directory.

        Examples:
            ```python
            experiment.download_artifacts(with_tree=True)
            ```

        Arguments:
            with_tree: If True, the artifacts will be downloaded in a tree structure.

        """
        if with_tree:
            self._setup_dirs()
            self._download_artifacts_with_tree_for_experiment()
        else:
            self._download_artifacts_without_tree_for_experiment()

    @exception_handler
    @beartype
    def _download_artifacts_with_tree_for_experiment(self):
        for artifact in self.list_artifacts():
            if artifact.name == "checkpoint-data-latest":  # pragma: no cover
                target_path = self.checkpoint_dir
            elif artifact.name == "checkpoint-index-latest":  # pragma: no cover
                target_path = self.checkpoint_dir
            elif artifact.name == "model-latest":  # pragma: no cover
                target_path = self.exported_model_dir
            elif artifact.name == "config":  # pragma: no cover
                target_path = self.config_dir
            else:
                target_path = self.base_dir

            artifact.download(target_path=target_path, force_replace=True)

    @exception_handler
    @beartype
    def _download_artifacts_without_tree_for_experiment(self):
        self._create_dir(self.base_dir)
        for artifact in self.list_artifacts():
            artifact.download(target_path=self.base_dir, force_replace=True)

    @exception_handler
    @beartype
    def set_base_experiment(self, base_experiment: "Experiment") -> None:
        """Set a base experiment to this experiment. Experiment must not be the same, and must be in the same project.

        If no "parameters" Log exists on this experiment, it will copy the base_experiment "parameters" (Log).
        If no "labelmap" Log exists on this experiment, it will copy the base_experiment "labelmap" (Log).
        If there are no Artifacts stored for this experiment, it will copy all the base_experiment (Artifact)
        If no base_model_version is set on this experiment, it will set the reference to base_experiment.model_version.

        Examples:
            ```python
            base_experiment = client.get_experiment_by_id("7b838580-fa66-4faa-bbc7-23cb70421fbd")
            my_experiment.set_base_experiment(base_experiment)
            ```
        Arguments:
            base_experiment (Experiment): An experiment to set as base experiment.
        """
        payload = {"base_experiment_id": base_experiment.id}
        self.connexion.post(
            f"/api/experiment/{self.id}/experiment", data=orjson.dumps(payload)
        )
        logger.info(f"{base_experiment} successfully set as base experiment to {self}")

    @exception_handler
    @beartype
    def set_base_model_version(
        self,
        base_model_version: "ModelVersion",
        attach_parameters: bool | None = None,
    ) -> None:
        """Set a base experiment to this experiment.

        If no "labelmap" Log exists on this experiment, it will copy the base_model_version "labelmap" (Log).
        If there are no Artifacts stored for this experiment, it will copy all the base_model_version (ModelFile)

        If attach_parameters is True, it will copy the base_model_version parameters inside a "parameters" Log.
        If attach_parameters is False, it will do nothing for parameters.
        If attach_parameters is None, it will copy the base_model_version parameters inside a "parameters" Log ONLY if there was no prior "parameters" Log on this experiment.

        Examples:
            ```python
            base_model_version = client.get_model_version_by_id("f91c4f26-e956-4a4f-ac75-d8371cf9928c")
            # This will copy parameters only if there were no parameters inside experiment.
            experiment.set_base_model_version(base_model_version)
            ```
        Arguments:
            base_model_version (ModelVersion): A model version to set as base model_version.
            attach_parameters (bool, optional): Define the behavior for parameters. Defaults to None.
        """
        payload = {
            "model_version_id": base_model_version.id,
            "do_attach_base_parameters": attach_parameters,
        }
        self.connexion.post(
            f"/api/experiment/{self.id}/model", data=orjson.dumps(payload)
        )
        logger.info(
            f"{base_model_version} successfully set as base model version to {self}"
        )

    @exception_handler
    @beartype
    @deprecated(
        deprecated_in="6.29.0",
        removed_in="7.0.0",
        details="This method should not be called anymore. Instead use set_base_model_version()",
    )
    def attach_model_version(
        self, model_version: ModelVersion, do_attach_base_parameters: bool = True
    ) -> None:
        """Be careful, this method is deprecated and will override base parameters of experiment if do_attach_base_parameters is not given"""
        return self.set_base_model_version(model_version, do_attach_base_parameters)

    @exception_handler
    @beartype
    def attach_dataset(self, name: str, dataset_version: DatasetVersion) -> None:
        """Attach a dataset version to this experiment.

        Retrieve or create a dataset version and attach it to this experiment.

        Examples:
            ```python
            foo_dataset = client.get_dataset("foo").get_version("first")
            my_experiment.attach_dataset("training", foo_dataset)
            ```
        Arguments:
            name (str): Name to label this attached dataset. Use it like a descriptor of the attachment.
            dataset_version (DatasetVersion): A dataset version to attach to the experiment.
        """
        payload = {"name": name, "dataset_version_id": dataset_version.id}
        self.connexion.post(
            f"/api/experiment/{self.id}/datasets", data=orjson.dumps(payload)
        )
        logger.info(f"{dataset_version} successfully attached to {self}")

    @exception_handler
    @beartype
    def detach_dataset(self, dataset_version: DatasetVersion) -> None:
        """Detach a dataset version from this experiment.

        Examples:
            ```python
            foo_dataset = client.get_dataset("foo").get_version("first")
            my_experiment.attach_dataset(foo_dataset)
            my_experiment.detach_dataset(foo_dataset)
            ```
        Arguments:
            dataset_version (DatasetVersion): A dataset version to attach to the experiment.
        """
        payload = {"ids": [dataset_version.id]}
        self.connexion.delete(
            f"/api/experiment/{self.id}/datasets", data=orjson.dumps(payload)
        )
        logger.info(f"{dataset_version} was successfully detached from {self}")

    @exception_handler
    @beartype
    @deprecated(
        deprecated_in="6.27.0",
        details="get_attached_dataset_versions can be used instead of this endpoint",
    )
    def list_attached_dataset_versions(self) -> list[DatasetVersion]:
        """Retrieve all dataset versions attached to this experiment

        Examples:
            ```python
            datasets = my_experiment.list_attached_dataset_versions()
            ```

        Returns:
            A list of (DatasetVersion) object attached to this experiment
        """
        return list(self.get_attached_dataset_versions().values())

    @exception_handler
    @beartype
    def get_attached_dataset_versions(self) -> dict[str, DatasetVersion]:
        """Retrieve all dataset versions attached to this experiment, in a dict where keys are aliases

        Examples:
            ```python
            datasets = my_experiment.get_attached_dataset_versions()
            ```

        Returns:
            A dict matching alias and (DatasetVersion) object attached to this experiment
        """
        r = self.connexion.get(f"/api/experiment/{self.id}/datasets").json()
        return {
            item["name"]: DatasetVersion(self.connexion, item["dataset_version"])
            for item in r["items"]
        }

    @exception_handler
    @beartype
    def get_dataset(self, name: str) -> DatasetVersion:
        """Retrieve the dataset version attached to this experiment with given name

        Examples:
            ```python
            dataset: Dataset = my_experiment.get_dataset('train')
            dataset_version: DatasetVersion = dataset.get_version("latest")
            pics = dataset.list_assets()
            annotations = dataset.list_annotations()
            ```

        Arguments:
            name (str): Name of the dataset version in the experiment

        Returns:
            A (DatasetVersion) object attached to this experiment
        """
        params = {"name": name}
        r = self.connexion.get(
            f"/api/experiment/{self.id}/datasets/find", params=params
        ).json()
        return DatasetVersion(self.connexion, r["dataset_version"])

    @exception_handler
    @beartype
    def run_train_test_split_on_dataset(
        self, name: str, prop: float = 0.8, random_seed: Any | None = None
    ) -> tuple[MultiAsset, MultiAsset, dict[str, list], dict[str, list], list[Label]]:
        """Run a train test split on a dataset attached to this experiment.

        Examples:
            ```python
            dataset: DatasetVersion = my_experiment.get_dataset('train')
            pics = dataset.list_assets()
            annotations = dataset.list_annotations()
            ```

        Arguments:
            name (str): Name of the dataset version in the experiment
            prop (float): Proportion of the dataset to use for training
            random_seed (int): Random seed to use for the split

        Returns:
            A tuple containing:
                - train_assets (MultiAsset): assets to use for training
                - eval_assets (MultiAsset): assets to use for evaluation
                - train_label_count (dict): number of assets per label in train set
                - eval_label_count (dict): number of assets per label in eval set
                - labels (list[Label]): list of labels in the dataset
        """
        dataset = self.get_dataset(name)
        (
            train_assets,
            eval_assets,
            train_label_count,
            eval_label_count,
            labels,
        ) = dataset.train_test_split(prop, random_seed)
        # TODO: Log those values
        return train_assets, eval_assets, train_label_count, eval_label_count, labels

    @exception_handler
    @beartype
    def add_evaluation(
        self,
        asset: Asset,
        add_type: str | AddEvaluationType = AddEvaluationType.REPLACE,
        rectangles: list[tuple[int, int, int, int, Label, float]] | None = None,
        polygons: list[tuple[list[list[int]], Label, float]] | None = None,
        classifications: list[tuple[Label, float]] | None = None,
        keypoints: list[tuple[list[list[int]], Label, float]] | None = None,
    ):
        """Add an evaluation of the asset by this experiment.

        By default, if given asset had already been evaluated, evaluation will be replaced.
        You can add different shapes but will only be able to compute evaluation metrics on one kind of inference type.

        Examples:
            ```python
            asset = dataset_version.find_asset(filename="asset-1.png")
            experiment.add_evaluation(asset, rectangles=[(10, 20, 30, 40, label_cat, 0.8), (50, 60, 20, 30, label_dog, 0.7)])
            job = experiment.compute_evaluations_metrics(InferenceType.OBJECT_DETECTION)
            job.wait_for_done()
            ```
        Arguments:
            asset (Asset): asset to add evaluation on
            add_type (str ou AddEvaluationType): replace or keep old evaluation, defaults to
            rectangles (optional): list of tuples representing rectangles with scores
            polygons  (optional): list of tuples representing polygons with scores
            classifications (optional): list of tuples representing classifications with scores
            keypoints (optional): list of tuples representing keypoints with scores
        """
        if (
            rectangles is None
            and classifications is None
            and polygons is None
            and keypoints is None
        ):
            raise ValueError(
                "Please give parameter 'rectangles', 'classifications', 'keypoints' or 'polygons'"
            )

        import_type = AddEvaluationType.validate(add_type)
        payload = {"import_type": import_type}
        payload_evaluation: dict[str, Any] = {"asset_id": asset.id}
        if rectangles is not None:
            payload_evaluation["rectangles"] = [
                {
                    "x": rectangle[0],
                    "y": rectangle[1],
                    "w": rectangle[2],
                    "h": rectangle[3],
                    "label_id": rectangle[4].id,
                    "score": rectangle[5],
                }
                for rectangle in rectangles
            ]

        if polygons is not None:
            payload_evaluation["polygons"] = [
                {"polygon": polygon[0], "label_id": polygon[1].id, "score": polygon[2]}
                for polygon in polygons
            ]
        if classifications is not None:
            payload_evaluation["classifications"] = [
                {"label_id": classification[0].id, "score": classification[1]}
                for classification in classifications
            ]
        if keypoints is not None:
            payload_evaluation["keypoints"] = [
                {
                    "keypoints": keypoint[0],
                    "label_id": keypoint[1].id,
                    "score": keypoint[2],
                }
                for keypoint in keypoints
            ]
        payload["evaluations"] = [payload_evaluation]

        self.connexion.post(
            f"/api/experiment/{self.id}/evaluations", data=orjson.dumps(payload)
        )

    @exception_handler
    @beartype
    def list_evaluations(
        self,
        limit: int | None = None,
        offset: int | None = None,
        page_size: int | None = None,
        order_by: list[str] | None = None,
        q: str | None = None,
    ) -> MultiEvaluation:
        """List evaluations of this experiment.
        It will retrieve all evaluations made by this experiment.
        You will then be able to manipulate them.

         Arguments:
            limit (int, optional): if given, will limit the number of evaluations returned
            offset (int, optional): if given, will return evaluations that would have been returned
                                    after this offset in given order
            page_size (int, optional): page size when returning evaluations paginated, can change performance
            order_by (list[str], optional): if not empty, will order evaluations by fields given in this parameter
            q (str, optional): if given, will try to apply query to filter evaluations

        Returns:
            A (MultiEvaluation)
        """
        if page_size:
            logger.warning("page_size is deprecated and not used anymore.")
        return TrainingService(self.connexion, self.id).list_evaluations(
            limit, offset, order_by, q
        )

    @exception_handler
    @beartype
    def compute_evaluations_metrics(
        self,
        inference_type: InferenceType,
        evaluations: (
            list[str | UUID] | list[Evaluation] | MultiEvaluation | None
        ) = None,
        worker: Worker | None = None,
        status: AnnotationStatus | None = None,
    ) -> Job:
        """Compute evaluation metrics across evaluations added to this experiment.
        Picsellia will compute coco metrics on each evaluation and compare to existing annotations.

        Workers parameter is deprecated and cannot be used anymore. It will be removed in 6.27

        Examples:
            ```python
            experiment.add_evaluation(rectangles=[(10, 20, 30, 40, label_cat, 0.8), (50, 60, 20, 30, label_dog, 0.7)])
            experiment.compute_evaluations_metrics(InferenceType.OBJECT_DETECTION)
            ```

        Arguments
            inference_type (InferenceType): Type of shapes pushed as evaluations.
            evaluations (MultiEvaluation or list of UUID or Evaluation, optional): Run coco evaluation on only given ids
            status (AnnotationStatus, optional): Existing annotations will be filtered to only retrieve those that have this status.

        Returns:
            A (Job) that you can wait for done.
        """

        payload = {"inference_type": inference_type}

        if evaluations:
            evaluation_ids = []
            for evaluation in evaluations:
                if isinstance(evaluation, Evaluation):
                    evaluation_ids.append(evaluation.id)
                else:
                    evaluation_ids.append(evaluation)

            payload["evaluation_ids"] = evaluation_ids

        if worker:
            logger.warning("worker is deprecated and should not be used anymore.")

        if status:
            payload["status"] = status

        r = self.connexion.post(
            f"/api/experiment/{self.id}/evaluate", data=orjson.dumps(payload)
        ).json()
        return Job(self.connexion, r, version=1)



import logging
from uuid import UUID

from beartype import beartype

from picsellia.colors import Colors
from picsellia.decorators import exception_handler
from picsellia.sdk.connexion import Connection
from picsellia.sdk.dao import Dao
from picsellia.types.enums import WorkerType
from picsellia.types.schemas import WorkerSchema

logger = logging.getLogger("picsellia")


class Worker(Dao):
    """
    This class is deprecated and should not be used anymore.
    """

    def __init__(self, connexion: Connection, data: dict, type: WorkerType | str):
        Dao.__init__(self, connexion, data)
        self.type = WorkerType.validate(type)

    @property
    def id(self) -> UUID:
        """Worker id is now the id of the user"""
        return self._user_id

    @property
    def username(self) -> str:
        """Username of the Worker"""
        return self._username

    @property
    def user_id(self) -> UUID:
        """id of the User"""
        return self._user_id

    def __str__(self):
        return (
            f"{Colors.UNDERLINE}Worker '{self.username}' of a {self.type} {Colors.ENDC}"
        )

    @exception_handler
    @beartype
    def sync(self) -> dict:
        logger.warning(
            "This method is not calling platform to refresh data. You should not call it"
        )
        return {"id": self.id, "user_id": self.user_id, "username": self.username}

    @exception_handler
    @beartype
    def refresh(self, data: dict) -> WorkerSchema:
        schema = WorkerSchema(**data)
        self._username = schema.username
        self._user_id = schema.user_id
        return schema

    @exception_handler
    @beartype
    def get_infos(self) -> dict:
        """Retrieve worker info

        Examples:
            ```python
            worker = my_dataset.list_workers()[0]
            print(worker.get_infos())
            ```

        Returns:
            A dict with data of the worker
        """
        return {"username": self.username}



import logging
from typing import Any
from uuid import UUID

import orjson
from beartype import beartype

from picsellia.colors import Colors
from picsellia.decorators import exception_handler
from picsellia.sdk.connexion import Connection
from picsellia.sdk.dao import Dao
from picsellia.sdk.dataset_version import DatasetVersion
from picsellia.sdk.experiment import Experiment
from picsellia.sdk.model_version import ModelVersion
from picsellia.types.schemas import ProjectSchema
from picsellia.utils import filter_payload

logger = logging.getLogger("picsellia")


class Project(Dao):
    def __init__(self, connexion: Connection, data: dict):
        Dao.__init__(self, connexion, data)

    @property
    def name(self) -> str:
        """Name of this (Project)"""
        return self._name

    def __str__(self):
        return f"{Colors.BOLD}Project '{self.name}' {Colors.ENDC} (id: {self.id})"

    @exception_handler
    @beartype
    def sync(self) -> dict:
        r = self.connexion.get(f"/api/project/{self.id}").json()
        self.refresh(r)
        return r

    @exception_handler
    @beartype
    def refresh(self, data: dict) -> ProjectSchema:
        schema = ProjectSchema(**data)
        self._name = schema.name
        return schema

    @exception_handler
    @beartype
    def get_resource_url_on_platform(self) -> str:
        """Get platform url of this resource.

        Examples:
            ```python
            print(project.get_resource_url_on_platform())
            >>> "https://app.picsellia.com/project/62cffb84-b92c-450c-bc37-8c4dd4d0f590"
            ```

        Returns:
            Url on Platform for this resource
        """

        return f"{self.connexion.host}/project/{self.id}"

    @exception_handler
    @beartype
    def list_experiments(
        self,
        limit: int | None = None,
        offset: int | None = None,
        order_by: list[str] | None = None,
    ) -> list[Experiment]:
        """Retrieve all experiments of this project

        Examples:
            ```python
            experiments = my_project.list_experiments()
            ```

        Arguments:
            limit (int, optional): Limit of experiments to retrieve. Defaults to None.
            offset (int, optional): Offset to start retrieving experiments. Defaults to None.
            order_by (list[str], optional): Order by fields. Defaults to None.

        Returns:
            A list of (Experiment) objects, that you can manipulate
        """
        params = {"limit": limit, "offset": offset, "order_by": order_by}
        params = filter_payload(params)
        response = self.connexion.get(
            f"/api/project/{self.id}/experiments", params=params
        ).json()
        return [Experiment(self.connexion, item) for item in response["items"]]

    @exception_handler
    @beartype
    def delete_all_experiments(self) -> None:
        """Delete all experiments of this project

        :warning: **DANGER ZONE**: Be very careful here!

        Examples:
            ```python
            my_project.delete_all_experiments()
            ```
        """
        experiment_ids = self.connexion.get(
            f"/api/project/{self.id}/experiments/ids"
        ).json()
        payload = {"ids": experiment_ids}
        self.connexion.delete(
            f"/api/project/{self.id}/experiments", data=orjson.dumps(payload)
        )
        logger.info(f"All experiments of {self} deleted.")

    @exception_handler
    @beartype
    def create_experiment(
        self,
        name: str,
        description: str | None = None,
        base_experiment: Experiment | None = None,
        base_model_version: ModelVersion | None = None,
        parameters: dict | None = None,
    ) -> Experiment:
        """Create an (Experiment) in this (Project).

        By giving parameters, it will create a (Log) parameters on this Experiment with the content of the dict.
        If base_experiment is provided, it will set the base_experiment on this Experiment.
        It will copy the labelmap and Artifacts.
        It might copy parameters of the base_experiment if no other parameters are given.
        If base_model_version is provided it will set the base_model_version on this Experiment.
        It will copy the labelmap and the model files, only if no base_experiment already fill them.
        It might copy the parameters of the base_model_version if no other parameters are given, only if no base_experiment already fill them.

        Examples:
            ```python
            base_model_version = client.get_model_version_by_id("fbddcb6e-16de-4226-9624-0737879af774")
            my_experiment = my_project.create_experiment(
                "test_experiment",
                description="This is a cool experiment",
                base_model_version=base_model_version,
            )
            ```
        Arguments:
            name (str, optional): Name of experiment. Defaults to None.
            description (str, optional): Description of experiment.
            base_experiment ((Experiment), optional): Base (Experiment) of this new experiment. Defaults to None.
            base_model_version ((ModelVersion), optional): Base (ModelVersion) of this new experiment. Defaults to None.
            parameters (dict, optional): Parameters
        Returns:
             A new (Experiment) of this project
        """
        payload: dict[str, Any] = {"name": name, "description": description or ""}
        if parameters is not None:
            payload["parameters"] = parameters

        if base_experiment is not None:
            payload["base_experiment_id"] = base_experiment.id

        if base_model_version is not None:
            payload["base_model_version_id"] = base_model_version.id

        r = self.connexion.post(
            f"/api/project/{self.id}/experiments", data=orjson.dumps(payload)
        ).json()
        experiment = Experiment(self.connexion, r)
        logger.info(f"{experiment} created")
        return experiment

    @exception_handler
    @beartype
    def update(
        self,
        name: str | None = None,
        description: str | None = None,
        private: bool | None = None,
    ) -> None:
        """Update a project with a new name or description

        Examples:
            ```python
            my_project.update(description="This is a cool project")
            ```

        Arguments:
            name (str, optional): New name of project. Defaults to None.
            description (str, optional): New description of project. Defaults to None.
            private (bool, optional): deprecated field.
        """
        payload = {"name": name, "description": description}

        if private is not None:
            logger.warning(
                "You cannot update privacy of a project anymore. This parameter will not be used"
            )

        # Filter None values
        filtered_payload = filter_payload(payload)
        r = self.connexion.patch(
            f"/api/project/{self.id}", data=orjson.dumps(filtered_payload)
        ).json()
        self.refresh(r)
        logger.info(f"{self} updated.")

    @exception_handler
    @beartype
    def delete(self) -> None:
        """Delete a project.

        :warning: **DANGER ZONE**: Be very careful here!

        It will delete the project and all experiments linked.

        Examples:
            ```python
            my_project.delete()
            ```
        """
        self.connexion.delete(f"/api/project/{self.id}")
        logger.info(f"{self} deleted.")

    @exception_handler
    @beartype
    def get_experiment(self, name: str) -> Experiment:
        """Retrieve an existing experiment by name.

        Examples:
            ```python
            my_experiment = my_project.get_experiment("test_experiment")
            ```
        Arguments:
            name (str, optional): Experiment's name.

        Raises:
            Exception: Experiment not found

        Returns:
            An (Experiment) object that you can manipulate
        """
        params = {"name": name}
        r = self.connexion.get(
            f"/api/project/{self.id}/experiments/find", params=params
        ).json()
        return Experiment(self.connexion, r)

    @exception_handler
    @beartype
    def get_experiment_by_id(self, id: UUID | str) -> Experiment:
        """Retrieve an existing experiment by id.

        Examples:
            ```python
            my_experiment = my_project.get_experiment_by_id("62cffb84-b92c-450c-bc37-8c4dd4d0f590")
            ```
        Arguments:
            id: Experiment's id.

        Raises:
            Exception: Experiment not found

        Returns:
            An (Experiment) object that you can manipulate
        """
        if isinstance(id, str):
            id = UUID(id)
        params = {"id": id}
        r = self.connexion.get(
            f"/api/project/{self.id}/experiments/find", params=params
        ).json()
        return Experiment(self.connexion, r)

    @exception_handler
    @beartype
    def attach_dataset(self, dataset_version: DatasetVersion) -> None:
        """Attach a dataset version to this project.

        Retrieve or create a dataset version and attach it to this project.

        Examples:
            ```python
            foo_dataset_version = client.get_dataset("foo").get_version("first")
            my_project.attach_dataset(foo_dataset_version)
            ```
        Arguments:
            dataset_version (DatasetVersion): A dataset version to attach to the project.
        """
        payload = {"dataset_version_id": dataset_version.id}
        self.connexion.post(
            f"/api/project/{self.id}/datasets", data=orjson.dumps(payload)
        )
        logger.info(f"{dataset_version} successfully attached to {self}")

    @exception_handler
    @beartype
    def detach_dataset(self, dataset_version: DatasetVersion) -> None:
        """Detach a dataset version from this project.

        Examples:
            ```python
            foo_dataset_version = client.get_dataset("foo").get_version("first")
            my_project.attach_dataset(foo_dataset_version)
            my_project.detach_dataset(foo_dataset_version)
            ```
        Arguments:
            dataset_version (DatasetVersion): A dataset version to attach to the project.
        """
        payload = {"ids": [dataset_version.id]}
        self.connexion.delete(
            f"/api/project/{self.id}/datasets", data=orjson.dumps(payload)
        )
        logger.info(
            f"{dataset_version} was successfully detached from this project {self}"
        )

    @exception_handler
    @beartype
    def list_dataset_versions(self) -> list[DatasetVersion]:
        """Retrieve all dataset versions attached to this project

        Examples:
            ```python
            datasets = my_project.list_dataset_versions()
            ```

        Returns:
            A list of (DatasetVersion) object attached to this project
        """
        r = self.connexion.get(f"/api/project/{self.id}/datasets").json()
        return [DatasetVersion(self.connexion, item) for item in r["items"]]



import logging
from typing import Any
from uuid import UUID

import orjson
from pydantic import Field, model_validator

from picsellia.sdk.asset import Asset
from picsellia.sdk.connexion import Connection
from picsellia.services.lister.default import AbstractItemLister, BaseFilter

logger = logging.getLogger(__name__)


class AssetFilter(BaseFilter):
    filenames: list[str] | None = Field(default=None, min_length=1)
    object_names: list[str] | None = Field(default=None, min_length=1)
    ids: list[str | UUID] | None = Field(default=None, min_length=1)
    filename_startswith: list[str] | None = Field(default=None, min_length=1)
    data_ids: list[str | UUID] | None = None
    assignment_status: str | None = None
    assignment_step_id: UUID | None = None
    assignment_user_id: UUID | None = None
    custom_metadata: dict | None = None

    @model_validator(mode="after")
    def check_query(self):
        with_filenames = self.filenames is not None
        with_object_names = self.object_names is not None
        with_ids = self.ids is not None
        with_data_ids = self.data_ids is not None
        with_filename_startswith = self.filename_startswith is not None

        if (
            not with_ids
            and not with_filenames
            and not with_object_names
            and not with_filename_startswith
            and not with_data_ids
        ):
            return self

        if self.limit:
            logger.warning(
                "limit cannot be used with parameters object_names, filenames, ids, data_ids, filename_startswith"
            )

        if self.offset:
            logger.warning(
                "offset cannot be used with parameters object_names, filenames, ids, data_ids, filename_startswith"
            )

        if (
            bool(self.ids)
            + bool(self.object_names)
            + bool(self.filenames)
            + bool(self.data_ids)
            + bool(self.filename_startswith)
            > 1
        ):
            raise ValueError(
                "You can only give one of filenames, object_names, ids, data_ids or filename_startswith"
            )

        return self

    def has_list_of_items(self) -> bool:
        return (
            self.object_names is not None
            or self.filenames is not None
            or self.ids is not None
            or self.data_ids is not None
            or self.filename_startswith is not None
        )


class AssetLister(AbstractItemLister[Asset, AssetFilter]):
    def __init__(self, connexion: Connection, dataset_version_id: UUID):
        super().__init__(connexion)
        self.dataset_version_id = dataset_version_id

    def _get_query_param_and_items_from_filters(
        self, filters: AssetFilter
    ) -> tuple[str, list]:
        if filters.ids:
            query_param = "ids"
            items = filters.ids
        elif filters.object_names:
            query_param = "object_names"
            items = filters.object_names
        elif filters.filenames:
            query_param = "filenames"
            items = filters.filenames
        elif filters.filename_startswith:
            query_param = "filename_startswith"
            items = filters.filename_startswith
        elif filters.data_ids:
            query_param = "data_ids"
            items = filters.data_ids
        else:  # pragma: no cover
            raise ValueError(
                "filenames, object_names, ids, filename_startswith or data_ids must be given"
            )
        return query_param, items

    def _get_other_params(self, filters: AssetFilter) -> dict[str, Any]:
        params = {}
        if filters.assignment_status:
            params["assignment_status"] = filters.assignment_status
        if filters.assignment_step_id:
            params["assignment_step_id"] = filters.assignment_step_id
        if filters.assignment_user_id:
            params["assignment_user_id"] = filters.assignment_user_id
        if filters.custom_metadata:
            params["custom_metadata"] = orjson.dumps(filters.custom_metadata)
        return params

    def _get_path(self) -> str:
        return f"/api/dataset/version/{self.dataset_version_id}/assets"

    def _build_item(self, item: dict) -> Asset:
        return Asset(self.connexion, self.dataset_version_id, item)



import logging
from operator import countOf
from pathlib import Path
from time import sleep
from uuid import UUID

import deprecation
import orjson
from beartype import beartype

from picsellia import exceptions as exceptions
from picsellia import pxl_multithreading as mlt
from picsellia.colors import Colors
from picsellia.decorators import exception_handler
from picsellia.exceptions import UnprocessableData, WaitingAttemptsTimeout
from picsellia.sdk.connexion import Connection
from picsellia.sdk.dao import Dao
from picsellia.sdk.data_projection import DataProjection
from picsellia.sdk.datasource import DataSource
from picsellia.sdk.downloadable import Downloadable
from picsellia.sdk.multi_object import MultiObject
from picsellia.sdk.tag import Tag, TagTarget
from picsellia.sdk.taggable import Taggable
from picsellia.types.enums import DataType, DataUploadStatus
from picsellia.types.schemas import DataSchema, ImageSchema, VideoSchema

logger = logging.getLogger("picsellia")


class Data(Dao, Downloadable, Taggable):
    def __init__(self, connexion: Connection, datalake_id: UUID, data: dict):
        Dao.__init__(self, connexion, data)
        Downloadable.__init__(self)
        Taggable.__init__(self, TagTarget.DATA)
        self._datalake_id = datalake_id
        self._embeddings = None

    def __str__(self):
        return f"{Colors.GREEN}Data{Colors.ENDC} object (id: {self.id})"

    @property
    def datalake_id(self) -> UUID:
        """UUID of (Datalake) where this (Data) is"""
        return self._datalake_id

    @property
    def object_name(self) -> str:
        """Object name of this (Data)"""
        return self._object_name

    @property
    def content_type(self) -> str:
        """Content type of this (Data)"""
        return self._content_type

    @property
    def filename(self) -> str:
        """Filename of this (Data)"""
        return self._filename

    @property
    def large(self) -> bool:
        """If true, this (Data) file is considered large"""
        return True

    @property
    def type(self) -> DataType:
        """Type of this (Data)"""
        return self._type

    @property
    def width(self) -> int:
        """Width of this (Data)"""
        if self._upload_status != DataUploadStatus.DONE:
            return 0

        return self._width

    @property
    def height(self) -> int:
        """Height of this (Data)"""
        if self._upload_status != DataUploadStatus.DONE:
            return 0

        return self._height

    @property
    @deprecation.deprecated(deprecated_in="6.13.0", removed_in="7.0.0")
    def duration(self) -> int:
        """This property is no longer supported"""
        return 0

    @property
    def metadata(self) -> dict | None:
        """Metadata of this Data. Can be None"""
        return self._metadata

    @property
    def custom_metadata(self) -> dict | None:
        """Custom metadata of this Data. Can be None"""
        return self._custom_metadata

    @property
    def upload_status(self) -> DataUploadStatus:
        """Status of upload of this Data. You can only use your data if this value is DONE."""
        return self._upload_status

    @property
    def embeddings(self) -> dict | None:
        """Embeddings of this Data. Can be None if data was not indexed"""
        if not self._embeddings:
            self._load_embeddings()
        return self._embeddings

    @exception_handler
    @beartype
    def refresh(self, data: dict):
        data_type = DataType.validate(data["type"])
        data_upload_status = DataUploadStatus.validate(data["upload_status"])

        if data_upload_status != DataUploadStatus.DONE:
            schema = DataSchema(**data)
        elif data_type == DataType.IMAGE:
            schema = ImageSchema(**data)
        elif data_type == DataType.VIDEO:
            schema = VideoSchema(**data)
        else:
            raise NotImplementedError("This data cannot be used at the moment.")

        # Downloadable properties
        self._object_name = schema.object_name
        self._filename = schema.filename
        self._url = schema.url
        self._metadata = schema.metadata
        self._custom_metadata = schema.custom_metadata
        self._upload_status = data_upload_status
        self._content_type = schema.content_type

        self._type = schema.type
        if data_upload_status == DataUploadStatus.DONE:
            self._height = schema.meta.height
            self._width = schema.meta.width

        return schema

    @exception_handler
    @beartype
    def sync(self) -> dict:
        r = self.connexion.get(f"/api/data/{self.id}").json()
        self.refresh(r)
        return r

    @exception_handler
    @beartype
    def reset_url(self) -> str:
        """Reset url property of this Data by calling platform.

        Returns:
            A url as a string of this Data.
        """
        r = self.connexion.get(f"/api/data/{self.id}/presigned-url")
        self._url = r.json()["presigned_url"]
        return self._url

    def is_ready(self) -> bool:
        return self.upload_status not in [
            DataUploadStatus.PENDING,
            DataUploadStatus.COMPUTING,
        ]

    @exception_handler
    @beartype
    def wait_for_upload_done(
        self, blocking_time_increment: float = 1.0, attempts: int = 20
    ):
        attempt = 0
        while attempt < attempts:
            self.sync()
            if self.is_ready():
                break

            sleep(blocking_time_increment)
            attempt += 1
            logger.info(
                f"Waited for {blocking_time_increment * attempt}s, trying {attempts - attempt} times again in {blocking_time_increment}s"
            )

        if attempt >= attempts:
            raise WaitingAttemptsTimeout(
                f"Data is still being processed, but we waited for {blocking_time_increment * attempts}s."
                "Please wait a few more moment"
            )

        if self.upload_status == DataUploadStatus.ERROR:
            logger.error("This data could not be processed by our services")
            raise UnprocessableData(self)

    @exception_handler
    @beartype
    def get_tags(self) -> list[Tag]:
        """Retrieve the tags of your data.

        Examples:
            ```python
            tags = data.get_tags()
            assert tags[0].name == "bicycle"
            ```

        Returns:
            List of (Tag) objects.
        """
        r = self.sync()
        return [Tag(self.connexion, item) for item in r["tags"]]

    @exception_handler
    @beartype
    def get_datasource(self) -> DataSource | None:
        """Retrieve (DataSource) of this Data if it exists. Else, will return None.

        Examples:
            ```python
            data_source = data.get_datasource()
            assert data_source is None
            ```

        Returns:
            A (DataSource) object or None.
        """
        r = self.sync()
        if "data_source" not in r or r["data_source"] is None:
            return None

        return DataSource(self.connexion, r["data_source"])

    @exception_handler
    @beartype
    def delete(self) -> None:
        """Delete data and remove it from datalake.

        :warning: **DANGER ZONE**: Be very careful here!

        Remove this data from datalake, and all assets linked to this data.

        Examples:
            ```python
            data.delete()
            ```
        """
        response = self.connexion.delete(f"/api/data/{self.id}")
        assert response.status_code == 204
        logger.info(f"1 data (id: {self.id}) deleted from datalake {self.datalake_id}.")

    @exception_handler
    @beartype
    def update_metadata(self, metadata: dict | None) -> None:
        """This method will update metadata of your (Data).
        If you want to update location of your Data, you must update both longitude and latitude at the same time.
        """
        payload = {"metadata": metadata}
        r = self.connexion.patch(
            f"/api/data/{self.id}", data=orjson.dumps(payload)
        ).json()
        self.refresh(r)
        logger.info(f"Metadata of {self.id} updated.")

    @exception_handler
    @beartype
    def update_custom_metadata(self, custom_metadata: dict | None):
        """This method will update custom_metadata of your (Data)."""
        payload = {"custom_metadata": custom_metadata}
        r = self.connexion.patch(
            f"/api/data/{self.id}/custom-metadata", data=orjson.dumps(payload)
        ).json()
        self._custom_metadata = r["custom_metadata"]
        logger.info(f"Metadata of {self.id} updated.")

    def _load_embeddings(self) -> dict | None:
        r = self.connexion.get(
            f"/api/visual-search/datalake/{self.datalake_id}/points/scroll",
            params={
                "with_vector": True,
                "with_payload": False,
                "has_error": False,
                "limit": 1,
            },
        ).json()
        if "points" not in r or len(r["points"]) == 0 or "vector" not in r["points"][0]:
            return None

        self._embeddings = r["points"][0]["vector"]
        return self._embeddings

    @exception_handler
    @beartype
    def list_projections(self, type: str | None = None) -> list[DataProjection]:
        """List (DataProjection) of this (Data).

        Examples:
            ```python
            projections = data.list_projections()
            projections[0].download()
            ```
        Returns:
            a list of (DataProjection)
        """
        if type is not None:
            logging.warning(
                "'type' parameter is deprecated and will be removed in future versions. "
            )

        r = self.connexion.get(f"/api/data/{self.id}/projections").json()
        return [DataProjection(self.connexion, self.id, item) for item in r["items"]]


class MultiData(MultiObject[Data], Taggable):
    @beartype
    def __init__(self, connexion: Connection, datalake_id: UUID, items: list[Data]):
        MultiObject.__init__(self, connexion, items)
        Taggable.__init__(self, TagTarget.DATA)
        self.datalake_id = datalake_id

    def __str__(self) -> str:
        return f"{Colors.GREEN}MultiData for datalake {self.datalake_id} {Colors.ENDC}, size: {len(self)}"

    def __getitem__(self, key) -> "Data | MultiData":
        if isinstance(key, slice):
            indices = range(*key.indices(len(self.items)))
            data = [self.items[i] for i in indices]
            return MultiData(self.connexion, self.datalake_id, data)
        return self.items[key]

    @beartype
    def __add__(self, other) -> "MultiData":
        self.assert_same_connexion(other)
        items = self.items.copy()
        if isinstance(other, MultiData):
            items.extend(other.items.copy())
        elif isinstance(other, Data):
            items.append(other)
        else:
            raise exceptions.BadRequestError("You can't add these two objects")

        return MultiData(self.connexion, self.datalake_id, items)

    @beartype
    def __iadd__(self, other) -> "MultiData":
        self.assert_same_connexion(other)

        if isinstance(other, MultiData):
            self.extend(other.items.copy())
        elif isinstance(other, Data):
            self.append(other)
        else:
            raise exceptions.BadRequestError("You can't add these two objects")

        return self

    def copy(self) -> "MultiData":
        return MultiData(self.connexion, self.datalake_id, self.items.copy())

    @exception_handler
    @beartype
    def split(self, ratio: float) -> tuple["MultiData", "MultiData"]:
        s = round(ratio * len(self.items))
        return MultiData(self.connexion, self.datalake_id, self.items[:s]), MultiData(
            self.connexion, self.datalake_id, self.items[s:]
        )

    @exception_handler
    @beartype
    def delete(self) -> None:
        """Delete a bunch of data and remove them from datalake.

        :warning: **DANGER ZONE**: Be very careful here!

        Remove a bunch of data from datalake, and all assets linked to each data.

        Examples:
            ```python
            whole_data = datalake.list_data(limit=3)
            whole_data.delete()
            ```
        """
        payload = {"ids": self.ids}
        self.connexion.delete(
            f"/api/datalake/{self.datalake_id}/datas",
            data=orjson.dumps(payload),
        )
        logger.info(f"{len(self.items)} data deleted from datalake {self.datalake_id}.")

    @exception_handler
    @beartype
    def download(
        self,
        target_path: str | Path = "./",
        force_replace: bool = False,
        max_workers: int | None = None,
        use_id: bool = False,
    ) -> None:
        """Download this multi data in given target path


        Examples:
            ```python
            bunch_of_data = client.get_datalake().list_data(limit=25)
            bunch_of_data.download('./downloads/')
            ```
        Arguments:
            target_path (str or Path, optional): Target path where to download. Defaults to './'.
            force_replace: (bool, optional): Replace an existing file if exists. Defaults to False.
            max_workers (int, optional): Number of max workers used to download. Defaults to os.cpu_count() + 4.
            use_id (bool, optional): If true, will download file with id and extension as file name. Defaults to False.
        """

        def download_one_data(item: Data):
            return item._do_download(target_path, force_replace, use_id=use_id)

        results = mlt.do_mlt_function(
            self.items, download_one_data, lambda item: item.id, max_workers=max_workers
        )
        downloaded = countOf(results.values(), True)

        logger.info(
            f"{downloaded} data downloaded (over {len(results)}) in directory {target_path}"
        )

    @exception_handler
    @beartype
    def wait_for_upload_done(
        self, blocking_time_increment: float = 1.0, attempts: int = 20
    ):
        errors = {}
        attempt = 0
        while attempt < attempts:
            still_pending = 0
            for data in self:
                if not data.is_ready():
                    data.sync()
                    if not data.is_ready():
                        still_pending += 1

                if (
                    data.upload_status == DataUploadStatus.ERROR
                    and data.filename not in errors
                ):
                    errors[data.filename] = data

            if still_pending == 0:
                break

            logger.info(
                f"{still_pending} are still being processed by our services, waiting for {blocking_time_increment}s "
                f"and retrying {attempts - attempt} times."
            )
            sleep(blocking_time_increment)
            attempt += 1

        if attempt >= attempts:
            raise WaitingAttemptsTimeout(
                f"Data is still being processed, but we waited for {blocking_time_increment * attempts}s. "
                "Please wait a few more moment"
            )

        if len(errors) > 0:
            filenames = ", ".join(list(errors.keys()))
            logger.error(
                f"Some data could not be processed by our services: {filenames}"
            )



from uuid import UUID

from orjson import orjson

from picsellia import exceptions
from picsellia.sdk.connexion import Connection
from picsellia.sdk.datasource import DataSource
from picsellia.utils import filter_payload


class DataSourceService:
    @classmethod
    def create_datasource(
        cls, connexion: Connection, organization_id: UUID, name: str
    ) -> DataSource:
        params = {"name": name}
        r = connexion.post(
            f"/api/organization/{organization_id}/datasources",
            data=orjson.dumps(params),
        ).json()
        return DataSource(connexion, r)

    @classmethod
    def get_datasource(
        cls, connexion: Connection, organization_id: UUID, name: str
    ) -> DataSource:
        params = {"name": name}
        r = connexion.get(
            f"/api/organization/{organization_id}/datasources/find",
            params=params,
        ).json()
        return DataSource(connexion, r)

    @classmethod
    def get_or_create_datasource(
        cls, connexion: Connection, organization_id: UUID, name: str
    ) -> DataSource:
        try:
            return cls.get_datasource(connexion, organization_id, name)
        except exceptions.ResourceNotFoundError:
            return cls.create_datasource(connexion, organization_id, name)

    @classmethod
    def list_datasources(
        cls,
        connexion: Connection,
        organization_id: UUID,
        limit: int | None = None,
        offset: int | None = None,
        order_by: list[str] | None = None,
    ) -> list[DataSource]:
        params = {"limit": limit, "offset": offset, "order_by": order_by}
        params = filter_payload(params)
        r = connexion.get(
            f"/api/organization/{organization_id}/datasources",
            params=params,
        ).json()
        return [DataSource(connexion, item) for item in r["items"]]



import logging
import mimetypes
import os
from pathlib import Path
from uuid import UUID

import requests
from requests import Session

from picsellia import exceptions
from picsellia.services.upload.adapters import (
    AzureUploadAdapter,
    GoogleMultipartUploadAdapter,
    S3MultipartUploadAdapter,
    S3UploadAdapter,
    UploadAdapter,
)
from picsellia.services.upload.platform import PlatformAdapter

logger = logging.getLogger("picsellia")

LARGE_FILE_SIZE = 25 * 1024 * 1024
DEFAULT_TIMEOUT = 30


class FileUploader:
    def __init__(self, connector_id: UUID, session: Session, host: str, headers: dict):
        self.platform_adapter = PlatformAdapter(connector_id, session, host, headers)

    def upload(
        self, object_name: str, path: str | Path
    ) -> tuple[requests.Response, bool, str]:
        is_large, content_type = self._read_file(path)

        if is_large:
            upload = self._upload_large_file(object_name, path, content_type)
        else:
            upload = self._upload_small_file(object_name, path, content_type)

        return upload, is_large, content_type

    def _upload_large_file(
        self,
        object_name: str,
        path: str | Path,
        content_type: str,
    ):
        try:
            response = self.platform_adapter.init_multipart_upload(
                object_name, content_type
            ).json()
            adapter = self._build_multipart_adapter(response)
        except exceptions.BadRequestError:
            logger.warning(
                "This file is large but it is impossible to use multipart for this upload."
                "Trying to upload in only one upload."
            )
            return self._upload_small_file(object_name, path, content_type)

        return adapter.upload(object_name, path, content_type)

    def _upload_small_file(self, object_name: str, path: str | Path, content_type: str):
        response = self.platform_adapter.generate_presigned_url(
            object_name, content_type
        ).json()
        adapter = self._build_adapter(response)
        return adapter.upload(object_name, path, content_type)

    @staticmethod
    def _read_file(path: str | Path) -> tuple[bool, str]:
        if not os.path.isfile(path):
            raise exceptions.FileNotFoundException(f"{path} not found")

        is_large = Path(path).stat().st_size > LARGE_FILE_SIZE
        content_type = mimetypes.guess_type(path, strict=False)[0]
        if content_type is None:
            content_type = "application/octet-stream"

        return is_large, content_type

    def _build_adapter(self, response: dict) -> UploadAdapter:
        if (
            "client_type" not in response
            or "presigned_url_data" not in response
            or "url" not in response["presigned_url_data"]
            or "fields" not in response["presigned_url_data"]
        ):
            raise exceptions.DistantStorageError(
                "Platform could not generate a presigned url to upload this file."
            )

        client_type = response["client_type"].lower()
        url = response["presigned_url_data"]["url"]
        fields = response["presigned_url_data"]["fields"]
        if client_type in ["aws", "minio", "google"]:
            return S3UploadAdapter(self.platform_adapter, url, fields)
        elif client_type == "azure":
            return AzureUploadAdapter(self.platform_adapter, url, fields)
        else:
            raise exceptions.BadRequestError(
                f"This version of SDK cannot upload on {client_type} connector."
            )

    def _build_multipart_adapter(self, response: dict) -> UploadAdapter:
        if "client_type" not in response:
            raise exceptions.BadRequestError(
                "Platform cannot accept a multipart upload for this file."
            )

        client_type = response["client_type"].lower()
        if client_type in ["aws", "minio"]:
            upload_id = response["upload_id"]
            return S3MultipartUploadAdapter(self.platform_adapter, upload_id)
        elif client_type == "google":
            url = response["url"]
            return GoogleMultipartUploadAdapter(self.platform_adapter, url)
        else:
            raise exceptions.BadRequestError(
                f"This version of SDK cannot upload on {client_type} connector."
            )



import logging
from uuid import UUID

import orjson
from beartype import beartype

from picsellia.colors import Colors
from picsellia.decorators import exception_handler
from picsellia.sdk.connexion import Connection
from picsellia.sdk.dao import Dao
from picsellia.sdk.label import Label
from picsellia.types.schemas import RectangleSchema
from picsellia.utils import filter_payload

logger = logging.getLogger("picsellia")


class Rectangle(Dao):
    def __init__(
        self,
        connexion: Connection,
        dataset_version_id: UUID,
        annotation_id: UUID,
        data: dict,
    ) -> None:
        self._dataset_version_id = dataset_version_id
        self._annotation_id = annotation_id
        Dao.__init__(self, connexion, data)

    @property
    def annotation_id(self) -> UUID:
        """UUID of the (Annotation) holding this (Rectangle)"""
        return self._annotation_id

    @property
    def x(self) -> int:
        """Coordinates x of this (Rectangle)"""
        return self._x

    @property
    def y(self) -> int:
        """Coordinates y of this (Rectangle)"""
        return self._y

    @property
    def w(self) -> int:
        """Width of this (Rectangle)"""
        return self._w

    @property
    def h(self) -> int:
        """Height of this (Rectangle)"""
        return self._h

    @property
    def text(self) -> str | None:
        """OCR text of this (Rectangle)"""
        return self._text

    @property
    def label(self) -> Label:
        """(Label) of this (Rectangle)"""
        return self._label

    def __str__(self):
        return f"{Colors.BLUE}Rectangle (x:{self.x},y:{self.y},w:{self.w},h:{self.h}) with label {self.label.name} on annotation {self.annotation_id} {Colors.ENDC} (id: {self.id})"

    @exception_handler
    @beartype
    def sync(self) -> dict:
        r = self.connexion.get(f"/api/rectangle/{self.id}").json()
        self.refresh(r)
        return r

    @exception_handler
    @beartype
    def refresh(self, data: dict) -> RectangleSchema:
        schema = RectangleSchema(**data)
        self._x = schema.x
        self._y = schema.y
        self._w = schema.w
        self._h = schema.h
        self._text = schema.text
        self._label = Label(
            self.connexion, self._dataset_version_id, schema.label.model_dump()
        )
        return schema

    @exception_handler
    @beartype
    def update(
        self,
        x: int | None = None,
        y: int | None = None,
        w: int | None = None,
        h: int | None = None,
        label: Label | None = None,
        text: str | None = None,
    ) -> None:
        """Update this rectangle with new coordinates or new label.

        Examples:
            ```python
            rect.update(x=10, label=label_car)
            ```

        Arguments:
            x (int, optional): New x coordinate of this (Rectangle). Defaults to None.
            y (int, optional): New y coordinate of this (Rectangle). Defaults to None.
            w (int, optional): New width of this (Rectangle). Defaults to None.
            h (int, optional): New height of this (Rectangle). Defaults to None.
            label (Label, optional): New label of this (Rectangle). Defaults to None.
            text (str, optional): New ocr text of this (Rectangle). Defaults to None.
        """
        payload = {"x": x, "y": y, "w": w, "h": h, "text": text}
        filtered_payload = filter_payload(payload)
        if label is not None:
            filtered_payload["label_id"] = label.id
        r = self.connexion.patch(
            f"/api/rectangle/{self.id}", data=orjson.dumps(filtered_payload)
        ).json()
        self.refresh(r)
        logger.info(f"{self} updated")

    @exception_handler
    @beartype
    def delete(self) -> None:
        """Delete this rectangle from the platform.

        :warning: **DANGER ZONE**: Be very careful here!

        Examples:
            ```python
            rect.delete()
            ```
        """
        self.connexion.delete(f"/api/rectangle/{self.id}")
        logger.info(f"{self} deleted.")



import abc
import logging
from datetime import date
from typing import ClassVar, Generic, TypeVar
from uuid import UUID

from beartype import beartype
from orjson import orjson

from picsellia.decorators import exception_handler
from picsellia.sdk.campaign.abstract_step import AbstractCampaignStep
from picsellia.sdk.dao import Dao
from picsellia.sdk.job import Job
from picsellia.sdk.user import User
from picsellia.sdk.worker import Worker
from picsellia.services.campaign import parse_user_weights_to_assignees
from picsellia.types.enums import CampaignStepType, WorkerType
from picsellia.utils import filter_payload

logger = logging.getLogger("picsellia")

TCampaignStep = TypeVar("TCampaignStep", bound=AbstractCampaignStep)


class AbstractCampaign(Dao, abc.ABC, Generic[TCampaignStep]):
    _base_path: ClassVar[str]
    _allowed_step_types: ClassVar[list[CampaignStepType]]
    _worker_type: ClassVar[WorkerType]

    @exception_handler
    @beartype
    def sync(self) -> dict:
        r = self.connexion.get(f"{self._base_path}/{self.id}").json()
        self.refresh(r)
        return r

    @exception_handler
    @beartype
    def close(self) -> None:
        """Close a campaign.

        :warning: **DANGER ZONE**: Be very careful here!

        It will close this campaign. The campaign will be locked and changes won't be allowed anymore.

        A closed dataset campaign CANNOT be reopened.
        If you need to reopen it, we suggest you fork the dataset version first and create a new campaign.

        A closed deployment campaign can be reopened.

        Examples:
            ```python
            foo_campaign.close()
            ```
        """
        self.connexion.post(f"{self._base_path}/{self.id}/close")
        logger.info(f"{self} closed")

    @exception_handler
    @beartype
    def delete(self) -> None:
        """Delete a campaign.

        :warning: **DANGER ZONE**: Be very careful here!

        It will remove this campaign from our database, all tasks and assignments will be removed.
        This will not delete annotations, reviews and assets.

        Examples:
            ```python
            foo_campaign.delete()
            ```
        """
        self.connexion.delete(f"{self._base_path}/{self.id}")
        logger.info(f"{self} deleted")

    def _update(
        self,
        description: str | None = None,
        instructions_object_name: str | None = None,
        instructions_text: str | None = None,
        end_date: date | None = None,
        auto_add_new_assets: bool | None = None,
        auto_close_on_completion: bool | None = None,
        default_entry_step: TCampaignStep | None = None,
        annotated_entry_step: TCampaignStep | None = None,
    ):
        if default_entry_step and default_entry_step.campaign_id != self.id:
            raise ValueError("Given default entry step must be in this campaign")

        if annotated_entry_step and annotated_entry_step.campaign_id != self.id:
            raise ValueError("Given annotated entry step must be in this campaign")

        payload = {
            "description": description,
            "instructions_object_name": instructions_object_name,
            "instructions_text": instructions_text,
            "end_date": end_date,
            "auto_add_new_assets": auto_add_new_assets,
            "auto_close_on_completion": auto_close_on_completion,
        }

        if default_entry_step:
            payload["default_entry_step_id"] = default_entry_step.id

        if annotated_entry_step:
            payload["annotated_entry_step_id"] = annotated_entry_step.id

        filtered_payload = filter_payload(payload)
        r = self.connexion.patch(
            f"{self._base_path}/{self.id}", data=orjson.dumps(filtered_payload)
        ).json()
        self.refresh(r)
        logger.info(f"{self} updated")

    @exception_handler
    @beartype
    def add_step(
        self,
        name: str,
        type: CampaignStepType | str,
        description: str | None = None,
        assignees: list[tuple[User | UUID | Worker, float | int]] | None = None,
        order: int | None = None,
        sample_rate: float | None = None,
    ) -> TCampaignStep:
        """Add a step on this campaign.

        `assignees` changed signature in 6.24, we now except a User or a user UUID.
        Worker will still be accepted until 6.26

        Examples:
            ```python
            users = foo_dataset.list_users()
            # In first step, 2 over 3 annotation task will be assigned to user 1, third one is going to user 0
            foo_campaign.add_step(
                name="annotation-step",
                type="ANNOTATION",
                description="annotation step",
                assignees=[(users[0], 1.0), (users[1], 2.0)]
            )
            # In second step, all review task will be assigned to user 2
            foo_campaign.add_step(
                name="review-step",
                type="REVIEW",
                description="review step",
                assignees=[(users[2], 1)]
            )
            ```
        Arguments:
            name (str): name of the step
            type (str or CampaignStepType): Type of the step: can be ANNOTATION or REVIEW
            description (str, optional): Description of the step. Defaults to None.
            assignees (list of tuple of (user or UUID), float, optional): Can be used to assign users to this step.
                Defaults to None.
            order (int, optional): Index where to insert the step in the workflow, the step will be appended at the end
                if nothing is specified. Defaults to None.
            sample_rate (float, optional): sample rate of this step. Defaults to None.
        Returns:
            dict, data of this step
        """
        campaign_step_type = CampaignStepType.validate(type)
        if campaign_step_type not in self._allowed_step_types:
            raise ValueError(
                f"Campaign step type must be in {self._allowed_step_types}"
            )

        payload = {
            "name": name,
            "type": campaign_step_type,
            "description": description,
            "order": order,
            "sample_rate": sample_rate,
        }
        if assignees:
            payload["assignees"] = parse_user_weights_to_assignees(assignees)

        r = self.connexion.post(
            f"{self._base_path}/{self.id}/steps",
            data=orjson.dumps(payload),
        ).json()
        logger.info(f"Step {name} has been added to this campaign")
        return self._build_step(r)

    @abc.abstractmethod
    def _build_step(self, data: dict):
        pass

    @exception_handler
    @beartype
    def list_steps(self) -> list[TCampaignStep]:
        """List all the steps of this campaign.

        Returns:
            a list of steps of this campaign as python dict
        """
        r = self.connexion.get(f"{self._base_path}/{self.id}/steps").json()
        return [self._build_step(item) for item in r["items"]]

    @exception_handler
    @beartype
    def _sync_assets(
        self, existing_annotations_step_id: str | UUID | None = None
    ) -> Job:
        r = self.connexion.get(
            f"{self._base_path}/{self.id}/assignments/sync/available"
        ).json()
        if "is_sync_available" in r and r["is_sync_available"] is False:
            reason = r["reason"]
            raise ValueError(f"Sync not available because {reason}")

        item_count = r["unsynced_item_count"]

        payload = {}
        if existing_annotations_step_id:
            payload["existing_annotations_step_id"] = existing_annotations_step_id

        r = self.connexion.post(
            f"{self._base_path}/{self.id}/assignments/sync",
            data=orjson.dumps(payload),
        ).json()
        self.refresh(r["campaign"])

        logger.info(
            f"Sync job is starting, {item_count} assets will be added to the campaign."
        )
        return Job(self.connexion, r["job"], version=2)



import logging
from uuid import UUID

from picsellia.colors import Colors
from picsellia.sdk.campaign.abstract_step import AbstractCampaignStep
from picsellia.sdk.connexion import Connection
from picsellia.sdk.dao import Dao

logger = logging.getLogger("picsellia")


class AnnotationCampaignStep(AbstractCampaignStep):
    _base_path = "/api/campaigns/annotation/steps"

    def __init__(self, connexion: Connection, data: dict, campaign_id: UUID):
        Dao.__init__(self, connexion, data)
        self._campaign_id = campaign_id

    def __str__(self):
        return f"{Colors.GREEN}Annotation Campaign Step{Colors.ENDC} for campaign {self.campaign_id} (id: {self.id})"

    @property
    def campaign_id(self) -> UUID:
        """UUID of the (AnnotationCampaign) of this step"""
        return self._campaign_id



import abc
import os
from io import BufferedReader
from typing import BinaryIO

import requests

from picsellia.decorators import retry
from picsellia.services.upload.platform import PlatformAdapter

CHUNK_SIZE = 25 * 1024 * 1024
DEFAULT_TIMEOUT = 30


class UploadAdapter(abc.ABC):
    def __init__(self, platform_adapter: PlatformAdapter):
        self.platform_adapter = platform_adapter
        self.session = self.platform_adapter.session

    @abc.abstractmethod
    def upload(self, object_name: str, path: str, content_type: str):
        pass


class S3UploadAdapter(UploadAdapter):
    def __init__(self, platform_adapter: PlatformAdapter, url: str, fields: dict):
        super().__init__(platform_adapter)
        self.url = url
        self.fields = fields

    @retry((requests.ReadTimeout, requests.ConnectionError))
    def upload(
        self, object_name: str, path: str, content_type: str
    ) -> requests.Response:
        with open(path, "rb") as file:
            url = self.url
            data = {**self.fields, "Content-Type": content_type}
            response = self.session.post(
                url=url,
                data=data,
                files={"file": (object_name, file)},
                timeout=DEFAULT_TIMEOUT,
            )
            response.raise_for_status()
            return response


class AzureUploadAdapter(UploadAdapter):
    def __init__(self, platform_adapter: PlatformAdapter, url: str, fields: dict):
        super().__init__(platform_adapter)
        self.url = url
        self.fields = fields

    @retry((requests.ReadTimeout, requests.ConnectionError))
    def upload(
        self, object_name: str, path: str, content_type: str
    ) -> requests.Response:
        with open(path, "rb") as file:
            url = self.url
            headers = {
                **self.fields,
                "Content-Type": content_type,
                "x-ms-blob-type": "BlockBlob",
            }
            response = self.session.put(
                url=url,
                data=file.read(),
                headers=headers,
                timeout=DEFAULT_TIMEOUT,
            )
            response.raise_for_status()
            return response


class S3MultipartUploadAdapter(UploadAdapter):
    def __init__(self, platform_adapter: PlatformAdapter, upload_id: str):
        super().__init__(platform_adapter)
        self.upload_id = upload_id

    def upload(
        self, object_name: str, path: str, content_type: str
    ) -> requests.Response:
        file_size = os.path.getsize(path)
        chunk_count = int(file_size / CHUNK_SIZE) + 1
        parts = []
        with open(path, "rb") as file:
            for part in range(1, chunk_count + 1):
                etag = self._upload_part(object_name, part, file)
                parts.append({"ETag": etag, "PartNumber": part})

        additional_payload = {
            "upload_id": self.upload_id,
            "parts": parts,
        }
        return self.platform_adapter.complete_part_upload(
            object_name, additional_payload
        )

    def _upload_part(
        self,
        object_name: str,
        part: int,
        file: BufferedReader | BinaryIO,
    ):
        additional_payload = {
            "upload_id": self.upload_id,
            "part_no": part,
        }
        url = self.platform_adapter.generate_part_presigned_url(
            object_name, additional_payload
        ).json()["url"]
        file_data = file.read(CHUNK_SIZE)
        response = self._put_part_with_retry(url, file_data)
        response.raise_for_status()
        return response.headers["ETag"]

    @retry((requests.ReadTimeout, requests.ConnectionError))
    def _put_part_with_retry(self, url: str, file_data: bytes):
        return self.session.put(url, data=file_data, timeout=DEFAULT_TIMEOUT)


class GoogleMultipartUploadAdapter(UploadAdapter):
    def __init__(self, platform_adapter: PlatformAdapter, url: str):
        super().__init__(platform_adapter)
        self.url = url

    def upload(
        self, object_name: str, path: str, content_type: str
    ) -> requests.Response:
        file_size = os.path.getsize(path)
        with open(path, "rb") as file:
            start = 0
            while start < file_size:
                chunk = file.read(CHUNK_SIZE)
                end = min(start + len(chunk) - 1, file_size - 1)
                content_length = end - start + 1
                headers = {
                    "Content-Length": str(content_length),
                    "Content-Range": f"bytes {start}-{end}/{file_size}",
                }
                response = self._put_part_with_retry(chunk, headers)
                response.raise_for_status()

                start = end + 1

            return response

    @retry((requests.ReadTimeout, requests.ConnectionError))
    def _put_part_with_retry(self, chunk: bytes, headers: dict):
        return self.session.put(
            url=self.url,
            data=chunk,
            headers=headers,
            timeout=DEFAULT_TIMEOUT,
        )



from typing import Generic, TypeVar
from uuid import UUID

from beartype import beartype

from picsellia import exceptions as exceptions
from picsellia.colors import Colors
from picsellia.sdk.connexion import Connection
from picsellia.sdk.dao import Dao

T = TypeVar("T", bound=Dao)


class MultiObject(Generic[T]):
    @beartype
    def __init__(self, connexion: Connection, items: list[T]) -> None:
        self._connexion = connexion

        if not items:
            raise exceptions.NoDataError("A MultiObject can't be empty")

        self.items: list[T] = items

    @property
    def connexion(self) -> Connection:
        return self._connexion

    @property
    def ids(self) -> list[UUID]:
        return [obj.id for obj in self.items]

    @beartype
    def __str__(self) -> str:
        return f"{Colors.GREEN}MultiObject{Colors.ENDC} object, size: {len(self.items)}"

    @beartype
    def __repr__(self) -> str:
        return self.__str__()

    @beartype
    def __eq__(self, other) -> bool:
        if isinstance(other, list):
            if len(other) != len(self.items):
                return False

            for item in self.items:
                if item not in other:
                    return False

            return True

        if isinstance(other, MultiObject):
            self.assert_same_connexion(other)
            if len(self.items) != len(other.items):
                return False

            for item in self.items:
                if item not in other.items:
                    return False

            return True

        return False

    @beartype
    def __len__(self) -> int:
        return len(self.items)

    @beartype
    def assert_enough_items(self) -> None:
        if len(self.items) < 2:
            raise exceptions.NoDataError(
                "You can't remove from this list the last item. A MultiObject can't be empty"
            )

    @beartype
    def assert_same_connexion(self, other: "Dao | MultiObject") -> None:
        if self.connexion != other.connexion:
            raise exceptions.BadRequestError(
                "Objects that you manipulate does not come from the same client."
            )

    @beartype
    def __delitem__(self, i: int) -> None:
        self.assert_enough_items()
        del self.items[i]

    @beartype
    def __setitem__(self, i: int, v: T) -> None:
        self.assert_same_connexion(v)
        self.items[i] = v

    @beartype
    def append(self, x: T) -> None:
        self.assert_same_connexion(x)
        self.items.append(x)

    @beartype
    def extend(self, objects: "T | MultiObject | list[T]") -> None:
        # objects cannot be typed MultiObject[T] because:
        # beartype.roar.BeartypeDecorHintForwardRefException: Forward reference 'MultiObject[T]' not valid Python attribute name.
        if isinstance(objects, Dao):
            self.assert_same_connexion(objects)
        elif isinstance(objects, MultiObject) or isinstance(objects, list):
            for x in objects:
                self.assert_same_connexion(x)
        else:
            raise TypeError(objects)
        self.items.extend(objects)

    @beartype
    def insert(self, i: int, x: T) -> None:
        self.assert_same_connexion(x)
        self.items.insert(i, x)

    @beartype
    def remove(self, x: T) -> None:
        self.assert_enough_items()
        self.items.remove(x)

    @beartype
    def pop(self, i: int | None = None) -> T:
        self.assert_enough_items()
        if i:
            return self.items.pop(i)
        else:
            return self.items.pop()

    @beartype
    def clear(self) -> None:
        raise exceptions.NoDataError("You can't clear a MultiObject.")

    @beartype
    def index(self, x: T, start: int | None = None, end: int | None = None) -> int:
        if start:
            if end:
                return self.items.index(x, start, end)
            else:
                return self.items.index(x, start)
        else:
            return self.items.index(x)

    @beartype
    def count(self, x: T) -> int:
        return self.items.count(x)

    @beartype
    def sort(self, key=None, reverse=False) -> None:
        self.items.sort(key=key, reverse=reverse)

    @beartype
    def reverse(self) -> None:
        self.items.reverse()

    @beartype
    def copy(self) -> "MultiObject":
        raise NotImplementedError("Not implemented. Please contact support")



import logging
from typing import Any
from uuid import UUID

import orjson

from picsellia.sdk.connexion import Connection
from picsellia.sdk.predicted_asset import PredictedAsset
from picsellia.services.lister.default import AbstractItemLister, BaseFilter

logger = logging.getLogger(__name__)


class PredictedAssetFilter(BaseFilter):
    assignment_status: str | None = None
    assignment_step_id: UUID | None = None
    assignment_user_id: UUID | None = None
    custom_metadata: dict | None = None

    def has_list_of_items(self) -> bool:
        return False


class PredictedAssetLister(AbstractItemLister[PredictedAsset, PredictedAssetFilter]):
    def __init__(self, connexion: Connection, deployment_id: UUID):
        super().__init__(connexion)
        self.deployment_id = deployment_id

    def _get_query_param_and_items_from_filters(
        self, filters: PredictedAssetFilter
    ) -> tuple[str, list]:
        raise NotImplementedError()

    def _get_other_params(self, filters: PredictedAssetFilter) -> dict[str, Any]:
        params = {}
        if filters.assignment_status:
            params["assignment_status"] = filters.assignment_status
        if filters.assignment_step_id:
            params["assignment_step_id"] = filters.assignment_step_id
        if filters.assignment_user_id:
            params["assignment_user_id"] = filters.assignment_user_id
        if filters.custom_metadata:
            params["custom_metadata"] = orjson.dumps(filters.custom_metadata)
        return params

    def _get_path(self) -> str:
        return f"/api/deployment/{self.deployment_id}/predictedassets"

    def _build_item(self, item: dict) -> PredictedAsset:
        return PredictedAsset(self.connexion, self.deployment_id, item)
